/*
** Module   : NameMappingWorker.java
** Abstract : maps Progress source names to Java class names
**
** Copyright (c) 2006-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20060201   @24183 Created initial version. Builds up maps
**                           which it uses to initialize and operate
**                           the SourceNameMapper runtime class.
** 002 GES 20060313   @25047 Some helpers to facilitate incremental name
**                           map processing (where only a subset of files
**                           can be processed at a time but they can
**                           still resolve classnames for files that were
**                           previously converted.
** 003 GES 20060319   @25113 Added helpers for external procedure
**                           signature processing.
** 004 GES 20101111          Pass conversion-time cfg values to convertName().
** 005 GES 20111213          resolveJavaClass returns null instead of throwing
**                           exception if the class was not found (4GL allows
**                           missing program name for RUN statements).
*/

package com.goldencode.p2j.convert;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.uast.*;
import com.goldencode.p2j.util.*;

/**
 * Gathers the mappings of Progress source names to Java class names and then
 * uses these maps to initialize and operate the <code>SourceNameMapper</code> 
 * runtime class.  This allows <code>RUN FILENAME</code> constructs to be  
 * mapped into the correct target classname.
 * <p>
 * This class only gathers the minimum data required to operate the
 * <code>convertName</code> method of the <code>SourceNameMapper</code>
 * class.  This minimizes the work and code needed.  In the future, it
 * would be easily possible to gather the method level data which would
 * allow additional Progress procedure/function to Java method mapping.
 */
public class NameMappingWorker
extends AbstractPatternWorker
{
   /** Maps Progress source file names to Java class names. */
   private Map p2jmap = new HashMap();
   
   /** Maps Progress source file names to external procedure signature. */
   private Map sigmap = new HashMap();
   
   /** List of the directories in the PROPATH of the legacy system. */
   private String[] propath = null; 
   
   /** Base path for all Java packages. */
   private String pkgroot = null;
   
   /** Base path for 4GL code. */
   private String proroot = null;
   
   /** Name element separator used for the Progress file name data. */
   private String fileSep = null; 
   
   /** Path separator used for the Progress file name data. */
   private String pathSep = null;

   /** Case sensitivity flag for the Progress file name data. */
   private boolean caseSens; 
   
   /** Flag to remember if we have initialized the source name mapper. */
   private boolean initialized = false; 
   
   /**
    * Default constructor which defines the symbol libraries to be registered.
    */
   public NameMappingWorker()
   {
      super();
      setLibrary(new Library());      
      
      // query base config values
      pkgroot  = Configuration.getParameter("pkgroot", "");
      proroot  = Configuration.getParameter("basepath", "./");
      fileSep  = Configuration.getParameter("file-separator", "/");
      pathSep  = Configuration.getParameter("path-separator", ":");
      propath  = Configuration.getParameter("propath", ".:").split(pathSep);
      
      String cs = Configuration.getParameter("case-sensitive", "true");
      caseSens = cs.equalsIgnoreCase("true");      
   }
      
   /**
    * Class whose instance is actually registered with the pattern engine's
    * symbol resolver as the primary library of name mapping user
    * functions.
    */
   public class Library
   {
      /**
       * Add a Progress to Java source file name entry to the mapping 
       * database.
       *
       * @param    progname
       *           Progress source file name in relative path form.
       * @param    javaname
       *           Java class name.
       */
      public void addMapping(String progname, String javaname)
      {
         progname = (caseSens ? progname : progname.toLowerCase()); 
         p2jmap.put(progname, javaname);
      }
      
      /**
       * Add the external procedure's parameter signature to the signature
       * database.
       *
       * @param    progname
       *           Progress source file name in relative path form.
       * @param    list
       *           List of parameter class names (each entry is a wrapper
       *           type) in proper parameter order.
       */
      public void addSignature(String progname, ArrayList list)
      {
         progname = (caseSens ? progname : progname.toLowerCase()); 
         sigmap.put(progname, list);
      }
      
      /**
       * Determine if a Progress to Java source file name mapping exists in 
       * the database.
       *
       * @param    progname
       *           Progress source file name in relative path form.
       *
       * @return   <code>true</code> if the mapping exists.
       */
      public boolean isMapped(String progname)
      { 
         return (p2jmap.get(progname) != null);
      }
      
      /**
       * Determine if a Progress external procedure has a <code>void</code>
       * signature (has no parameters).
       *
       * @param    progname
       *           Progress source file name in relative path form.
       *
       * @return   <code>true</code> if the signature mapping DOES NOT exist
       *           or if the mapping exists but the length is 0.
       */
      public boolean isVoid(String progname)
      { 
         ArrayList list = (ArrayList) sigmap.get(progname);
         return (list == null || list.size() == 0);
      }
      
      /**
       * Determine the type of the given Progress external procedure's
       * parameter at a given index.
       *
       * @param    progname
       *           Progress source file name in relative path form.
       *
       * @return   The wrapper type of the parameter specified or
       *           <code>null</code> if the signature mapping DOES NOT exist
       *           or if the mapping exists but there is no parameter at the
       *           given index.
       */
      public String getParameterType(String progname, int index)
      { 
         ArrayList list = (ArrayList) sigmap.get(progname);
         
         String type = null;
         
         if (list != null && index >= 0 && index < list.size())
         {
            type = (String) list.get(index);
         }
         
         return type;
      }
      
      /**
       * Obtain the parameter list for the given Progress external procedure.
       *
       * @param    progname
       *           Progress source file name in relative path form.
       *
       * @return   The list of parameters or <code>null</code> if the
       *           signature mapping DOES NOT exist.
       */
      public ArrayList getSignature(String progname)
      { 
         return (ArrayList) sigmap.get(progname);
      }
      
      /**
       * Returns the number of mappings in the database.
       *
       * @return   The number of mappings.
       */
      public int size()
      { 
         return p2jmap.size();
      }
      
      /**
       * Save the name mapping data into the given file using serialization. 
       *
       * @param    filename
       *           The file in which to write the name mappings.
       */
      public void saveMappings(String filename)
      throws FileNotFoundException,
             IOException
      {
         FileOutputStream   fos = new FileOutputStream(filename, false);
         ObjectOutputStream oos = new ObjectOutputStream(fos);
         
         oos.writeObject(p2jmap);
         oos.writeObject(sigmap);
         oos.close();
      }
      
      /**
       * Restore the name mapping data from the given file using
       * serialization. 
       *
       * @param    filename
       *           The file from which to read the name mappings.
       */
      public void restoreMappings(String filename)
      throws FileNotFoundException,
             IOException,
             ClassNotFoundException
      {
         FileInputStream   fis = new FileInputStream(filename);
         ObjectInputStream ois = new ObjectInputStream(fis);
         
         p2jmap = (HashMap) ois.readObject();
         sigmap = (HashMap) ois.readObject();
         ois.close();
      }
      
      /**
       * Force the external initialization of the 
       * <code>SourceNameMapper</code> class such that it can be used
       * during conversion for resolving a Java class name from the
       * Progress source file name.  Note that only the method
       * <code>convertName</code> can work properly as only a subset of the
       * mapping data is being loaded into the <code>SourceNameMapper</code>
       * class.
       */
      public void initSourceNameMapper()
      {
         SourceNameMapper.overwriteMappingData(p2jmap,
                                               null,
                                               null,
                                               null,
                                               propath,
                                               pkgroot,
                                               proroot,
                                               fileSep,
                                               pathSep,
                                               caseSens);
      }
      
      /**
       * Use the <code>SourceNameMapper</code> class to convert the given
       * Progress external procedure name into a proper Java class name.
       * This method will return the fully qualified class name of
       * the target class IF that class is not in the same package.
       *
       * @param    progname
       *           The Progress external procedure name.
       * @param    pkgname
       *           The package name of the currently processing file's
       *           converted class.  This name MUST not end in a ".".
       *
       * @return   The Java class name.
       */
      public String resolveJavaClass(String progname, String pkgname)
      {
         // first time in, initialize the source name mapper
         if (!initialized)
         {
            initSourceNameMapper();
            initialized = true;
         }
         
         String full = SourceNameMapper.convertName(progname);
         
         if (full == null)
         {
            return null;
         }
         
         // find the possible last separator
         int len = pkgname.length() + 1;
         
         // same package?
         if (full.startsWith(pkgname) && full.indexOf(".", len) == -1) 
         {  
            // prune off the package name and return
            return full.substring(len);
         }
         
         return full;
      }
   }
}
