/*
** Module   : Namespace.java
** Abstract : A searchable Progress schema namespace
**
** Copyright (c) 2004-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 ECF 20041223   @19172 Created initial version.
** 002 ECF 20050118   @19339 Added nodes() method to allow read-only
**                           iteration over list of nodes in namespace,
**                           for use by SchemaDictionary.
** 003 ECF 20050203   @19566 Added a parameter to the find method to
**                           specify whether search must find an exact
**                           match or an abbreviated one. This change was
**                           made to allow SchemaDictionary to mimic the
**                           actual symbol resolution performed by
**                           Progress. Progress looks first for an exact
**                           match in all scopes. Only if this fails are
**                           all scopes searched again for an abbreviated
**                           match.
** 004 ECF 20050207   @19627 Added the contains method and refactored the
**                           various find methods, such that contains can
**                           use the same underlying search algorithm.
** 005 ECF 20050209   @19733 Partially reverted search algorithm to
**                           pre-003 (@19566) behavior, based on further
**                           testing against live Progress source code.
**                           However, retained the "exact" flag in find
**                           method, which now allows an override of the
**                           dual search (exact, then abbreviated match)
**                           behavior to force exact match only.
** 006 ECF 20050221   @19900 Moved empty iterator implementation from
**                           nodes() method into a standalone class:
*                            com.goldencode.p2j.util.EmptyIterator.
** 007 ECF 20050413   @20706 Changed NameNode constructor called in find
**                           method.
** 008 GES 20070402   @32703 Honor the preferred flag for a name node to
**                           disambiguate an ambiguous name conflict. If
**                           1 (and only 1) of the ambiguous name nodes
**                           is marked as preferred, then match with that
**                           node. Added a copy constructor.
** 009 GES 20071130   @36189 Allow nodes to be marked as maskable. Callers
**                           obtaining an iterator can then choose if
**                           the maskable nodes are or are not included
**                           in the iterator.
** 010 ECF 20080730   @39271 Reduced memory footprint. Lazily initialize
**                           collections of nodes and maskables. Use
**                           smaller default sizes for these collections.
** 011 GES 20090421   @41831 Matched utility class package change.
** 012 ECF 20111217          Massive performance improvement: maintain sorting
**                           of nodes list as nodes are added, rather than
**                           using deferred sort.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.util.*;

/**
 * A namespace is used for unambiguous lookups of schema entity names. Objects
 * of this class are used by the {@link SchemaDictionary} to perform lookups
 * of schema entity (i.e., database, table, field) names within various
 * scopes.
 * <p>
 * A namespace contains a pool of {@link NameNode}s, each of which in turn
 * contains its own <code>Namespace</code> object. This hierarchy goes at most
 * three levels deep, reflecting the qualified name hierarchy represented by
 * the Progress language naming conventions for schema entity references:
 * <pre>
 *    database.table.field
 *    database.table
 *    table.field
 *    database
 *    table
 *    field
 * </pre>
 * where each part of a fully qualified name represents an entity which has
 * a one-to-many relationship with the entity to its right (if any), and a
 * one-to-one relationship with the entity to its left (if any).
 * <p>
 * Additional, global namespaces for lookups of unqualified database, table,
 * and field names are maintained by {@link SchemaDictionary}.
 * <p>
 * <code>Namespace</code> supports lookups of both unabbreviated and
 * abbreviated names. Whether abbreviation lookups are supported is determined
 * at construction. An exact match will always be preferred to an abbreviated
 * match. Only one exact match or one abbreviated match is permitted. If more
 * are detected, the search is considered ambiguous, and an exception will be
 * thrown by the {@link #findUnambiguously} method.
 */
public class Namespace
{
   /** Pool of name nodes comprising the namespace */
   private List<NameNode> nodes = null;
   
   /** Name nodes which can be excluded from the namespace iterator. */
   private Set<NameNode> maskable = null;
   
   /** Permit lookups using abbreviated names */
   private boolean abbreviations = false;
   
   /**
    * Constructor which sets abbreviated name lookup behavior.
    *
    * @param   abbreviations
    *          <code>true</code> to allow abbreviations, <code>false</code>
    *          to disallow.
    */
   public Namespace(boolean abbreviations)
   {
      this.abbreviations = abbreviations;
   }
   
   /**
    * Copy constructor which creates a separate and independent instance.
    *
    * @param   other
    *          The instance from which to pattern.
    */
   public Namespace(Namespace other)
   {
      this.abbreviations = other.abbreviations;
      
      if (other.nodes != null)
      {
         this.nodes = new ArrayList<NameNode>(other.nodes);
      }
      
      if (other.maskable != null)
      {
         this.maskable = new HashSet<NameNode>(other.maskable);
      }
   }
   
   /**
    * Add a name node to the namespace node pool at its naturally ordered
    * index (such that the pool is always sorted).
    *
    * @param   node
    *          Name node to be added to the pool.
    */
   public synchronized void add(NameNode node)
   {
      add(node, false);
   }
   
   /**
    * Add a name node to the namespace node pool at its naturally ordered
    * index (such that the pool is always sorted). Optionally mark the node as
    * one that can be excluded from iteration.
    *
    * @param   node
    *          Name node to be added to the pool.
    * @param   exclude
    *          <code>true</code> to mark this node as one that can be
    *          optionally excluded from an iterator.
    */
   public synchronized void add(NameNode node, boolean exclude)
   {
      if (nodes == null)
      {
         nodes = new ArrayList<NameNode>(1);
      }
      
      int index = Collections.binarySearch(nodes,
                                           node,
                                           NameNode.EXACT_COMPARATOR);
      if (index < 0)
      {
         index = -(index + 1);
      }
      
      nodes.add(index, node);
      
      if (exclude)
      {
         if (maskable == null)
         {
            maskable = new HashSet<NameNode>(4);
         }
         maskable.add(node);
      }
   }
   
   /**
    * Remove a name node from the namespace node pool and, if successful, set
    * a flag to indicate the list needs to be sorted. This method does not
    * actually re-sort the list (this is done lazily when a search is
    * performed).
    *
    * @param   node
    *          The node to be removed from the pool.
    *
    * @return  <code>true</code> if the node was removed; <code>false</code>
    *          if it was not present in the pool.
    */
   public synchronized boolean remove(NameNode node)
   {
      boolean success = false;
      
      if (nodes != null)
      {
         success = nodes.remove(node);
         
         // remove if it is there, we don't care about failures
         if (maskable != null)
         {
            maskable.remove(node);
         }
         
         if (success && nodes.size() == 0)
         {
            nodes = null;
         }
      }
      
      return success;
   }
   
   /**
    * Get read-only iterator over ALL name nodes in this namespace. List is
    * sorted before the iterator returns.
    *
    * @return  Fail-fast iterator over the name node list.
    *
    * @throws  ConcurrentModificationException
    *          if the backing node list is modified while this iterator is
    *          in use.
    */
   public synchronized Iterator<NameNode> nodes()
   {
      return nodes(false);
   }
   
   /**
    * Get read-only iterator over the name nodes in this namespace. List is
    * sorted before the iterator returns.
    *
    * @param   exclude
    *          <code>true</code> to exclude nodes (marked as maskable) from
    *          the iterator. <code>false</code> to return all contained nodes.
    *
    * @return  Fail-fast iterator over the name node list.
    *
    * @throws  ConcurrentModificationException
    *          if the backing node list is modified while this iterator is
    *          in use.
    */
   public synchronized Iterator<NameNode> nodes(boolean exclude)
   {
      if (nodes == null)
      {
         // Return a do-nothing iterator.
         return new EmptyIterator<NameNode>();
      }
      
      List<NameNode> subset = nodes;
      
      if (exclude && maskable != null)
      {
         subset = new ArrayList<NameNode>(nodes.size() - maskable.size());
         
         for (NameNode nn : nodes)
         {
            if (!maskable.contains(nn))
            {
               subset.add(nn);
            }
         }
      }
      
      return Collections.unmodifiableList(subset).iterator();
   }
   
   /**
    * Indicate whether this namespace contains the specified name node.
    *
    * @param   node
    *          Name node for which to test.
    *
    * @return  <code>true</code> if node is found, else <code>false</code>.
    */
   public synchronized boolean contains(NameNode node)
   {
      return (find(node, NameNode.EXACT_COMPARATOR) >= 0);
   }
   
   /**
    * Find a name node with the specified name in the pool of nodes. This
    * performs a lazy sort of the node pool list if it is currently unsorted.
    * This method attempts to find an exact match for the specified name in
    * this namespace. Then if not found and abbreviated matches are permitted
    * by this instance, it attempts to find an abbreviated match. If one (and
    * only one) match is found, it is returned.
    *
    * @param   name
    *          Name for which to search (may be an unambiguous abbreviation
    *          for the target node's name if this namespace instance permits
    *          abbreviation lookups). Name may match no more than one node in
    *          this namespace.
    * @param   exact
    *          <code>true</code> to require an exact match;
    *          <code>false</code> to allow an exact or an abbreviated name to
    *          match.
    *
    * @return  The matching {@link NameNode} object from the node pool, if an
    *          unambiguous match was found, or <code>null</code> if no match
    *          was found. Also returns <code>null</code> if the internal pool
    *          of name nodes has not yet been initialized.
    *
    * @throws  AmbiguousSchemaNameException
    *          if <code>name</code> represents an ambiguous search key (i.e.,
    *          matches more than one node in the pool).
    */
   public synchronized NameNode find(String name, boolean exact)
   throws AmbiguousSchemaNameException
   {
      // No contents to search through.
      if (nodes == null)
      {
         return null;
      }
      
      NameNode key = new NameNode(name);
      
      // Look for an exact match first.
      NameNode found = findUnambiguously(key, NameNode.EXACT_COMPARATOR);
      
      // If no exact match was found and abbreviations are allowed, look for
      // an abbreviation match.
      if (found == null && abbreviations && !exact)
      {
         found = findUnambiguously(key, NameNode.ABBREVIATION_COMPARATOR);
      }
      
      return found;
   }
   
   /**
    * Find the name node in the pool which matches the specified search key,
    * using <code>comparator</code> to perform a binary search. If a match
    * is found, test the nodes immediately before and after the found node
    * for matches using the same match criteria. If any are found, the
    * search key is determined to be ambiguous and an exception is thrown.
    *
    * @param   key
    *          Search key node used for comparison with nodes in the pool.
    *          Only the name stored in this key node is used in the search;
    *          other attributes are not considered and are therefore not
    *          necessary.
    * @param   comparator
    *          Comparator which is used to find a match in a binary search.
    *
    * @return  The matching <code>NameNode</code> object from the node pool,
    *          if an unambiguous match was found, or <code>null</code> if no
    *          match was found.
    *
    * @throws  AmbiguousSchemaNameException
    *          if the search key is ambiguous (i.e., matches more than one
    *          node in the pool).
    */
   private NameNode findUnambiguously(NameNode key,
                                      Comparator<NameNode> comparator)
   throws AmbiguousSchemaNameException
   {
      NameNode found = null;
      
      int index = find(key, comparator);
      if (index >= 0)
      {
         List<NameNode> matches = new LinkedList<NameNode>();
         matches.add(nodes.get(index));
         
         // Walk backward from index and check for duplicate matches.
         boolean hit = true;
         for (int i = index - 1; hit && i >= 0; i--)
         {
            NameNode next = nodes.get(i);
            hit = (comparator.compare(next, key) == 0);
            if (hit)
            {
               matches.add(0, next);
            }
         }
         
         // Walk forward from index and check for duplicate matches.
         hit = true;
         int size = nodes.size();
         for (int i = index + 1; hit && i < size; i++)
         {
            NameNode next = nodes.get(i);
            hit = (comparator.compare(next, key) == 0);
            if (hit)
            {
               matches.add(next);
            }
         }
         
         // If more than one match, this search is ambiguous; throw an
         // exception.
         int count = matches.size();
         if (count > 1)
         {
            // Try to disambiguate the name by checking if there is 1 (and
            // only 1) preferred node in the list. If so, return that node.
            found = findPreferred(matches);
            
            if (found == null)
            {
               // The preferred flag didn't help. Abort.
               throw new AmbiguousSchemaNameException(
                  "'" +
                  key.getName() +
                  "' is ambiguous; " +
                  count +
                  " matches found",
                  matches);
            }
         }
         else
         {
            found = nodes.get(index);
         }
      }
      
      return found;
   }
   
   /**
    * Returns the only match that is preferred or <code>null</code> if there
    * are no preferred matches or if there is more than 1 preferred match.
    *
    * @param   matches
    *          The list of all matches from which to find a preferred match.
    *
    * @return  The match that is preferred or <code>null</code> if there is
    *          no single match that is preferred.
    */
   private NameNode findPreferred(List<NameNode> matches)
   {
      Iterator<NameNode> iter = matches.iterator();
      
      NameNode found = null;
      
      while (iter.hasNext())
      {
         NameNode next = iter.next();
         
         if (next.isPreferred())
         {
            if (found == null)
            {
               // first preferred node
               found = next;
            }
            else
            {
               // there is more than 1 preferred node, we are done
               return null;
            }
         }
      }
      
      // if found is non-null, it must have been the only preferred match
      // in the list
      return found;
   }
   
   /**
    * The actual implementation of the find node algorithm, this method lazily
    * sorts all name nodes in this namespace into their natural order, then
    * performs a binary search using <code>comparator</code>. If the list of
    * nodes has not yet been initialized at the time this method is invoked,
    * it returns <code>-1</code> immediately.
    * <p>
    * See also the java.util.Collections binarySearch method description.
    *
    * @param   key
    *          A name node whose name matches the target node.
    * @param   comparator
    *          Comparator which directs the binary search algorithm.
    *
    * @return  The index of a matching name node within the internal list
    *          <code>nodes</code>, or <code>-1</code> to indicate that the
    *          search found no match.
    */
   private int find(NameNode key, Comparator<NameNode> comparator)
   {
      // No contents to search through.
      if (nodes == null)
      {
         return -1;
      }
      
      return Collections.binarySearch(nodes, key, comparator);
   }
}
