/*
** Module   : Browse.java
** Abstract : core browse widget implementation
**
** Copyright (c) 2005-2012, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ------------------Description-------------------
** 001 SVG 20050803   @21995 Initial version.
** 002 SIY 20050831   @22486 Refactoring (see @@22394-22409). 
** 003 SIY 20050907   @22623 Fixed use of @inheritDoc tag in javadoc
** 004 SIY 20050909   @22819 Added attribute container, added and updated 
**                           methods for interaction with BrowseColumn, 
**                           added support for SENSITIVE attribute, 
**                           implemented Validateable interface 
**                           (temporarily commented out).
** 005 SIY 20051205   @24198 Added necessary methods to support Progress 
**                           methods and attributes.
** 006 SIY 20060131   @24199 Use ThinClient special method instead of
**                           direct calls to Window.message().
** 007 SIY 20060209   @24487 Updated to accommodate changes in xxxConfig
**                           classes.
** 008 SIY 20060315   @25119 Major changes to complete BROWSE 
**                           implementation.
** 009 SIY 20060324   @25274 Fixes to make behavior match Progress 
**                           counterpart.
** 010 NVS 20060703   @27766 Reviewed the uses of requestFocus() and
**                           setFocus(). They seem to be safe as they are
**                           internal focus changes.
** 011 NVS 20060703   @27799 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 012 GES 20060713   @28001 Added support to honor the help attribute.
** 013 GES 20060808   @28493 Use int instead of long for row IDs. While
**                           this theoretically limits the maximum number 
**                           of rows per query, in practice the server
**                           side already has the same limitation AND it 
**                           is not practical to ever reach such limits
**                           in Progress code anyway. So we will take the
**                           small benefit of faster network performance
**                           over the capacity since the browse has a
**                           great potential to be a heavy consumer of
**                           network bandwidth.
** 014 SIY 20060824   @28844 Removed getP2JType() method.
** 015 SIY 20061025   @30637 Fixed size calculation and added support for
**                           column type override (@ fields).
** 016 SIY 20061027   @30757 Added refreshing of the browse on 
**                           VALUE-CHANGED event. Fixed drawing of empty
**                           browse.
** 017 SIY 20061102   @30939 Small change to encapsulate field.
** 018 SIY 20061105   @30982 Reposition browse to current row in the model.
** 019 SIY 20061114   @31361 Rewritten cache/navigation. Fixed insert mode
**                           handling and refreshing of row changes.
** 020 NVS 20061204   @31552 Switched to using a private widgetFocused
**                           flag instead of config.selected
** 021 ECF 20061204   @31579 Changed code to interpret a current row of -1
**                           to indicate a query off-end condition. A row
**                           should not be selected in this case.
** 022 ECF 20061206   @31588 Modified refreshStartRow(). This method now
**                           invalidates the browse instead of drawing it
**                           immediately.
** 023 ECF 20061207   @31621 Backed out #022 (@31588). This was causing a
**                           regression in that a reposition against the
**                           query was not correctly updating the browse
**                           view to reflect a new highlighted row.
** 024 SIY 20061208   @31628 refreshStartRow() now explicitly calls
**                           refresh().
** 025 SIY 20061208   @31654 Fix for initial viewport position and drawing
**                           of empty browse.
** 026 SIY 20061219   @31715 Adjusted refresh() behavior to make it similar
**                           to Progress.
** 027 SIY 20070315   @32418 Fixed page width calculation.
** 028 SIY 20070319   @32461 Set model current row in selectFocusedRow().
** 029 SIY 20070321   @32587 Fixed initialization of empty browse, fixes
**                           for layout calculations.
** 030 NVS 20070426   @33247 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 031 SIY 20070504   @33420 New approach in model size handling.
** 032 SIY 20070515   @33747 Fixed positioning.
** 033 SIY 20070730   @34723 More positioning fixes.
** 034 SIY 20070731   @34743 Fixed handling of browse/column help messages.
** 035 SIY 20070926   @35251 Fixed numeric fields alignment. 
**                           Ref: #35250
** 036 ECF 20070928   @35272 Fixed row updates. Removed nullable check in
**                           stopEdit(). Added internal, END-ERROR driven,
**                           row level undo processing to match Progress'
**                           behavior on failed field constraint checks.
** 037 ECF 20071011   @35447 Key handling fixes. Added handler for END
**                           function (move to last row). Fixed PAGE UP
**                           and PAGE DOWN handlers. Disallow negative row
**                           index in setCurrentRow(), except for special
**                           case of END handler.
** 038 ECF 20071030   @35631 Modified refresh(boolean). Added code to draw
**                           after refreshing the data cache. Modified
**                           refreshStartRow() to remove draw() call,
**                           since it is now redundant.
** 039 GES 20071227   @36593 Added hasFocusable() method.
** 040 SIY 20080121   @36873 Changed sending of VALUE-CHANGED. Attempt  
**                           to limit redrawing in some cases.
**                           Ref: #36871
** 041 SIY 20080128   @36920 More correct checking of conditions when 
**                           redraw is required.
**                           Ref: #36919
** 042 SVL 20080208   @36972 Force right-aligning when the column format
**                           is changed from 'number' to 'character'.
** 043 NVS 20080401   @37770 Added isFocusable() method implementation.
** 044 SIY 20080409   @37883 Fixed focus and keyboard and focus handling.
**                           Ref: #37887
** 045 NVS 20080414   @37982 ENTRY event is never consumed by Browse and
**                           may only be consumed by NO-APPLY triggers.
** 046 SIY 20080416   @37995 Postponed refreshing if frame is not visible.
**                           Ref: #37994
** 047 SIY 20080505   @38213 Fixed drawing of not focused BROWSE.
** 048 SIY 20080506   @38217 Force reinitialization of the underlying query
**                           when it is known that previously query
**                           returned other values. Some performance
**                           improvements.
**                           Ref: #38216
** 049 SIY 20080512   @38255 Fixed drawing of disabled widget.
**                           Removed #048, persistence fixes made it 
**                           redundant. More performance improvements.
** 050 SIY 20080521   @38362 Fixed keyboard handling for editing mode.
**                           Ref: #38355
** 051 SIY 20080617   @38796 Fixed initial display of empty browse.
** 052 SIY 20080623   @38925 Fixed drawing of special case when browse is
**                           displayed before query is assigned.
** 053 SIY 20080703   @39036 Fixed drawing of horizontal delimiter.
** 054 CA  20080901   @39644 Fixed startEdit: must show the help for the
**                           the focused widget.
** 055 SIY 20080903   @39695 More fixes in drawing of delimiters and
**                           highlighting.
** 056 CA  20080911   @39784 Bug fix in DataColumnRenderer.addData - the
**                           column and var type must be compared using a
**                           case-insensitive method.
** 057 SVL 20080925   @39941 Call to repaint() was added to
**                           refreshStartRow().
** 058 SIY 20081105   @40313 Added flag which tracks 'focus attractor'
**                           state. See #40284 for more details.
** 059 SIY 20081107   @40368 Rolled back #058, it is not necessary since
**                           more generic approach is discovered.
**                           See #40367 for more details.
** 060 CA  20081114   @40451 Changes due to new clipping implementation.
** 061 CA  20081203   @40757 When an event is received, clear the temp
**                           texts from the message area.
** 062 CA  20081208   @40821 Added ROW-DISPLAY event implementation.
** 063 CA  20081210   @40845 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 064 CA  20081211   @40861 Fixed a regression introduced by H062: after
**                           ROW-DISPLAY trigger invocation, reposition 
**                           the query to the previous 'scrollRow' value.
** 065 CA  20081212   @40884 Yet another regression about ROW-DISPLAY: the
**                           current row must be computed from scrollRow 
**                           and currRow.
** 066 CA  20081212   @40887 On CLOSE QUERY, reset the currRow field.
**                           triggerRowDisplayEvents: row is restored from
**                           the currRow value.
** 067 SIY 20090213   @41309 Added 'force' parameter into refreshStartRow
**                           signature.
** 068 NVS 20090213   @41312 Name changed for clearPutScreenInMessageArea
**                           method in ThinCLient.
** 069 SIY 20090216   @41322 Added tracking of highlighting state.
** 070 CA  20090403   @41728 Fix regression introduced by frame disabling
**                           on destroy - generate ROW-DISPLAY events only
**                           if frame is still alive.
** 071 SIY 20090413   @41784 Fix regression caused by changes in focus 
**                           management. Do not remove highlighting on 
**                           CHARVA focus change since it does not reflect
**                           P2J widget highlighting.
** 072 CA  20090423   @41926 Fixed a regression introduced by ROW-DISPLAY
**                           events - these events must be triggered only
**                           on initial draw or when the rows have scrolled.
**                           The problem was that, on each draw, the query was 
**                           repositioned (which might interfere with an
**                           ongoing CREATE/UPDATE operation).
** 073 SIY 20090902   @43824 Browse is always valid (just like other simple 
**                           components).
** 074 SIY 20090908   @43840 Do not redraw widget if parent frame is hidden.
** 075 SVL 20091011   @44133 Now we take into account the offset of the zero
**                           row of the rowDisplayed array (which equals
**                           dcStartRow).
** 076 SVL 20091029   @44287 Reworked logic into refresh(boolean) responsible
**                           for determination of the top row in a view.
** 077 SVL 20091105   @44315 Into refresh(boolean) do not set the current row
**                           as the start row in a view if the total number of
**                           rows is less than or equals view size.
** 078 SVL 20091119   @44441 Fixed repositioning for editable browse, fixed
**                           adjustment of current row on focus gaining event
**                           if the current row of the backing query row was
**                           changed because of GET statement, fixed
**                           browse:refresh() for the case of server-side
**                           update of rows.
** 079 EVL 20100430   @44839 Fix for incorrect help string with empty browser
**                           data.
** 080 SIY 20100619          Switching to own implementation of character UI.
** 081 SIY 20100804          Changes in imports, some refactorings.
** 082 SIY 20100822          Refactored AttributeContainer support.
** 083 SIY 20100015          locateFrame() is moved to UiUtils, renamed base
**                           class.
** 084 SIY 20101122          Switching from KeyEvent to KeyInput.
** 085 SIY 20101202          Minor change in constant name.
** 086 SIY 20110112          Changes related to new color handling scheme.
** 087 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 088 SIY 20111202          Refactored and moved from CHUI package.
** 089 SIY 20111221          Switched to widget factory.
** 090 CA  20121022          Added support for optional query and column-list
**                           clauses.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of Progress BROWSE widget.
 * 
 * TODO:  Do we handle options like NO-ASSIGN, NO-VALIDATE, and VALIDATE
 * functions?
 */
public abstract class Browse 
extends FixedSizeContainer
implements FocusListener,
           KeyListener
{
   /** The size of data cache in screen pages. */
   public static final int CACHE_SIZE = 20;
   
   /** A data model instance. */
   protected BrowseModel browseModel = null;

   /**
    * Is in-place edit can be enabled. This is indicated with '*' sign in
    * the leftmost column.
    */
   protected boolean editPossible = false;
   
   /** Zero-based index of current row. */
   protected int currRow = -1;
   
   /**
    * Index of current column. It is the column where the last edit is
    * performed.
    */
   protected int currColumn = 0;
   
   /**
    * An updated data for the current row. These data are passed to model if
    * focus leaves this row. If the array or some element in it is
    * <code>null</code> then no updates are available for the corresponding
    * cell.
    */
   protected BaseDataType[] currRowUpdates = null;
   
   /**
    * Indexes of selected rows in <code>Integer</code> form. Used only in
    * multiple selection mode. This mode is active if and only if this member
    * is not <code>null</code>.
    */
   protected TreeSet<Integer> selectedRowIndices = null;
   
   /** The data model index of the first row in the view window. */
   protected int scrollRow = 0;
   
   /**
    * The data model index of the scroll anchor column. Value <code>0</code>
    * means the absence of horizontal scrolling. Otherwise this column will
    * occupy the right part of the view window.
    */
   protected int scrollColumn = 0;
   
   /**
    * The height (in character rows) of the labels area not including the
    * separator line.
    */
   protected int ilLabelsHeight = 0;
   
   /** Number of character character columns in the displayable page. */
   protected int ilPageWidth = 0;
   
   /** Number of rows in the displayable page. */
   protected int ilPageRowsCount = 0;
   
   /**
    * Number of space columns on screen between the data area and the right
    * border.
    */
   protected int ilRightSpaceWidth = 0;
   
   /**
    * The minimal positive value of <code>scrollColumn</code>.
    */
   protected int ilFirstScrollCol = 0;
   
   /**
    * Cached rows from the data model. This is circular buffer containing
    * a sequential range of data rows.
    */
   protected BaseDataType[][] dcData = null;
   
   /**
    * The index in the data model of the first cache row.  
    */
   private int dcStartRow = 0;
   
   /** Number of rows in the data model. */
   protected int dcRowCount = 0;

   /** Separator line length */
   protected int separatorLen = 0;
   
   /**
    * Rendered columns labels area. This cache is reset when horizontal
    * scrolling is performed.
    */
   protected String[] rcLabels = null;
   
   /** Component attribute container. */
   protected BrowseConfig config = null;
   
   /** Draw labels. */
   protected boolean drawLabels = true;

   /** Help message for current column. */
   private String columnHelp = null;
   
   /** Waiting for END-ERROR key after edit mode validation error */
   private boolean awaitingEndError = false;

   /** Track postponed refresh requests. */
   protected boolean postponedRefresh = false;

   /** Track postponed start row reset requests. */
   private boolean postponedRowReset = false;

   /** Set to true if browse was drawn at least once. */
   protected boolean hadDraw = false;
   
   /** 
    * This variable is used to reproduce special drawing behavior for case
    * when query is initialized after browse is displayed first time. 
    */
   protected boolean pseudoFocus = false;

   /** Set to <code>true</code> when separators need to be underscored. */
   public boolean separatorQuirk = false;

   /** Keep track of each row display status. */
   private boolean[] rowDisplayed = null;

   /** The P2J color for each cell. */
   protected Color[][] colors = null;

   /** Protect ROW-DISPLAY events against recursion. */
   private boolean rowDisplayTriggered = false;

   /** Quirk to ignore a {@link #rowDisplayTriggered} set to ON case. */
   private boolean ignoreQuirk = false;

   /** Info about at which row started the last drawing. */
   protected int oldDisplayRow = -1;

   /**
    * Terminal color for current text. It is displayed in inverse colors.
    */
   public Color colorCurrent;

   /**
    * Terminal color for normal text in data area when separators are
    * active. It is displayed with underline.
    */
   protected Color colorHSep;

   /** Terminal attribute for normal text. */
   public Color colorCommon;

   /**
    * Constructs the Browse widget.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public Browse(ComponentConfig cfg)
   {
      config = new BrowseConfig(this, cfg);
      config.applyConfig(cfg);
      
      addKeyListener(this);
      addFocusListener(this);
   }

   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#setEnabled(boolean)
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      if (enabled && hadDraw)
         separatorQuirk  = true;

      if(enabled != isEnabled())
         pseudoFocus = false;

      if (enabled)
      {
         resetRowDisplayState();
      }

      super.setEnabled(enabled);
   }
   
   /**
    * Tracks the highlighted state  of the component.
    * 
    * @param  highlighted
    *         New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      if (highlighted != highlighted())
         pseudoFocus = false;
      
      super.setHighlighted(highlighted);
   }
   
   /**
    * @see com.goldencode.p2j.ui.client.widget.Widget#setVisible(boolean)
    */
   @Override
   public void setVisible(boolean visible)
   {
      super.setVisible(visible);
   }
   
   /**
    * Second stage of the initialization of the Browse widget which is
    * performed only when query is assigned.
    * 
    * @param   queryId
    *          Server side query ID.
    */
   public void setQuery(int queryId)
   {
      if (browseModel != null)
         return;

      browseModel = new ClientBrowseModel(queryId);

      BrowseColumnConfig[] cols = browseModel.getColumns();

      ilLabelsHeight = 0;
      ilPageRowsCount = config.getDown();
      
      int minHeight = config.getDown();
      int minWidth  = 4;

      if (config.isBox())
      {
         minWidth += 2;
         minHeight += 2;
      }
      
      editPossible = false;
      int firstEditable = -1;
      
      drawLabels = config.isLabels();
      
      for (int i = 0; i < cols.length; ++i)
      {
         if (!config.isLabels())
         {
            cols[i].setLabels(false);
            if (cols[i].isForceLabel())
               drawLabels = true;
         }
         
         ilLabelsHeight = Math.max(ilLabelsHeight, 
                                   cols[i].getLabelVar().length);
         if (!cols[i].isReadOnly())
         {
            editPossible = true;
            
            if (firstEditable == -1)
               firstEditable = i;
         }
         
         minWidth += cols[i].getWidthChars();
         
         if (i < (cols.length - 1))
            minWidth += 1;
      }
      
      if (drawLabels)
         minHeight += ilLabelsHeight + 1; //Headers
      
      if (config.getWidth() > 0)
         minWidth = config.getWidth();
      
      if (config.getWidthChars() == 0 && config.getHeightChars() == 0)
      {
         config.setHeightChars(minHeight);
         config.setWidthChars(minWidth);
      }
      
      setSize(config.getWidthChars(), config.getHeightChars());
      
      if (config.isMultiple())
         selectedRowIndices = new TreeSet<Integer>();

      // init cache
      int current = browseModel.getCurrentRow();
      
      initCache(current, current, true);

      if (editPossible && dcRowCount > 0)
      {
         if (selectedRowIndices != null)
         {
            selectedRowIndices.add(new Integer(1));
         }
         currColumn = firstEditable;
         
         startEdit();
      }
      else if (editPossible && firstEditable >= 0)
         config.setHelp(cols[firstEditable].getHelp());

      if (config.isMultiple())
         setSelectionMultiple(true);
   }

   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   public BaseDataType getValue()
   {
      // editing is allowed
      if (getActiveEditor() != null)
         return getActiveEditor().getValue(); 
         
      // editing is not allowed
      // TODO: change the approach
      // Code below may cause problems in GUI environment, because each 
      // validation will reset view to selected cell. This is not an issue
      // in TUI because there is no way to make current cell not in view port. 
      cellToView(currRow, currColumn);
      
      return getValue(currColumn);
   }
   
   /**
    * Method to force validation.
    * 
    * @return  <code>true</code> if validation test passed, <code>false</code>
    *          otherwise.
    */
   public boolean validateThis()
   {
      // commit changes
      config.setModified(false);
      
      return releaseCurrentRow();
   }
   
   /**
    * Check if value is modified since last commit.
    * 
    * @return  <code>true</code> if value was changed.
    */
   public boolean isModified()
   {
      return isCurrentRowModified() || config.isModified();
   }

   /**
    * Handle situation when <code>Browse</code> widget receives the focus.
    * <p>
    * When a help string is defined for the browse it will be displayed on
    * the status line.  This method also sets the selected attribute on. 
    * 
    * @param    event
    *           Reference to the event.
    */
   @Override
   public void onFocusGained(FocusEvent event)
   {
      setCurrentRow(currRow, true);

      if (columnHelp != null && config.getHelp() == null)
      {
         columnHelp = null;
      }
      else
      {
         ThinClient.setHelp(config.getHelp());
      }

      repaint();
   }
   
   /**
    * Handle situation when <code>Browse</code> widget loses the focus.
    * <p>
    * This method clears the focused flag.
    * 
    * @param    fe_
    *           Reference to the event.
    */
   @Override
   public void onFocusLost(FocusEvent fe_)
   {
      repaint();
   }
   
   /**
    * Undo editing operation results and reset everything to initial state.
    */
   public void undoValue()
   {
      FillIn editor = getActiveEditor();
      
      if (editor != null)
         stopEdit();
      
      config.setModified(false);
      currRowUpdates = null;
      
      if (editor != null)
         startEdit();
   }
   
   /**
    * Browse behaves like simple component in this regard.
    * 
    * @return  <code>true</code>.
    */
   public boolean valid()
   {
      return true;
   }

   /**
    * Requesting the focus to the <code>Browse</code>.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;

      setCurrentRow(currRow, true);

      super.requestFocus();
   }
   
   /**
    * Set READ-ONLY attribute.
    * 
    * @param   readOnly
    *          New value for the attribute.
    */
   public void setReadOnly(boolean readOnly)
   {
      if (browseModel == null)
         return;
         
      if (readOnly)
         stopEdit();
      else
         startEdit();
   }

   /**
    * Select all rows.
    */
   public void selectAll()
   {
      if (selectedRowIndices != null)
      {
         for (int i = 0; i < dcRowCount; i++)
         {
            Integer invIndex = new Integer(i);
            
            if (!selectedRowIndices.contains(invIndex))
               selectedRowIndices.add(invIndex);
         }
         draw();
         requestSync();
      }
   }
   
   /**
    * Remove all selections. 
    */
   public void deselectAll()
   {
      if (selectedRowIndices != null)
      {
         selectedRowIndices.clear();
         draw();
         requestSync();
      }
   }
   
   /**
    * Select row which is currently in focus. 
    */
   public void selectFocusedRow()
   {
      if (currRow < 0 || browseModel == null)
      {
         return;
      }
      
      if (selectedRowIndices != null)
      {
         Integer invIndex = new Integer(currRow);
         if (!selectedRowIndices.contains(invIndex))
         {
            selectedRowIndices.add(invIndex);
            draw();
            requestSync();
         }
      }

      browseModel.setCurrentRow(currRow);
   }
   
   /**
    * Renders the screen value into a cell. This is helper method for the
    * render classes.
    * 
    * @param   screen
    *          Screen value.
    * @param   col
    *          Column to render.
    * @param   colWidth
    *          The full width for formatting.
    * @param   scroll
    *          Scrolling correction. Means the number of characters which
    *          should be erased from the left (if positive) or from the right
    *          (if negative). This scroll differs from the scroll in
    *          <code>DisplayFormat.formatScreenValue()</code> because it
    *          means the scroll of the whole browse column not the scroll of
    *          screen value inside the column.
    * 
    * @return  Rendered screen value.
    */
   protected static String renderCell(String screen, BrowseColumnConfig col, 
                                     int colWidth, int scroll)
   {
      boolean rightAlign = col.getDisplayFormat().isRightPositioning() ||
                           col.isRightPositioning();
      
      // adjust alignment for numbers
      if (col.getDisplayFormat() instanceof NumberFormat)
         rightAlign = true;
      
      if (col.isRightPositioning())
         screen = screen.trim();
      
      String display = DisplayFormat.formatScreenValue(screen, colWidth, 0, 
                                                       rightAlign, rightAlign);
      if (scroll > 0)
      {
         // scroll by truncating from the left
         display = display.substring(scroll);
      }
      else
      {
         display = display.substring(0, display.length() + scroll);
      }

      return display;
   }

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public ComponentConfig config()
   {
      return config;
   }
   
   /**
    * Sets position of terminal's cursor. Redirects this positioning to the
    * in-place editor if it is present. Otherwise does nothing because BROWSE
    * itself has no purpose for the hardware cursor.
    */
   @Override
   public abstract void drawCaret();

   /**
    * Draws the widget: border, header, selection marks and line, internal
    * FillIn etc.
    */
   @Override
   public abstract void draw();

   /**
    * Sets new widget's size.
    *
    * @param width
    *        The width component of the new size.
    * @param height
    *        The height component of the new size.
    *
    * @throws IllegalArgumentException
    *         When new size is not large enough.
    */
   @Override
   public void setSize(int width, int height)
   {
      if (browseModel == null || width == 0 || height == 0)
         return;

      // mark columns
      ilPageWidth = width - 2;
      ilPageRowsCount = height;

      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      int colWidth = 0;
      
      for (int i = 0; i < cols.length; i++)
      {
         colWidth += cols[i].getWidthChars() + 1;
      }
      
      if (cols.length > 0)
      {
         // the last column has no vertical separator
         colWidth--;
      }
      
      if (config.isBox())
      {
         // if no columns specified, use the explicit width set by the user.
         ilPageWidth = cols.length > 0 
                          ? Math.min(colWidth, width - 4)
                          : width - 4;
         ilPageRowsCount -= 2;
      }

      if (drawLabels)
      {
         ilPageRowsCount -= ilLabelsHeight + 1;
      }
      
      if (ilPageWidth < 1 || ilPageRowsCount < 2)
      {
         throw new IllegalArgumentException
            ("Too little browse size: " + width + "x" + height +
             ". It implies the data area size: " + ilPageWidth + "x" +
             ilPageRowsCount);
      }

      ilRightSpaceWidth = ilPageWidth - colWidth;
      
      if (ilRightSpaceWidth <= 0)
      {
         ilRightSpaceWidth = 0;
      }
      
      ilFirstScrollCol = cols.length;
      int w = -1;
      
      for (int i = 0; i < cols.length; ++i)
      {
         w += cols[i].getWidthChars() + 1;
         
         if (w > ilPageWidth)
         {
            ilFirstScrollCol = i;
            break;
         }
      }

      separatorLen = ilPageWidth - (config.isBox() ? 0 : 2);

      super.setSize(width, height);
   }

   /**
    * This component will receive an input focus when Tab or Shift-Tab
    * is pressed.
    * 
    * @return <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns the same as <code>getDefaultFocus()</code> because focus placement
    * inside BROWSE can't be switched from initial component until subcomponent
    * creation or deletion.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget currentFocus()
   {
      return defaultFocus();
   }

   /**
    * Returns this component directly if there is no in-place editor.
    * Otherwise returns this editor.
    * 
    * @see com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget defaultFocus()
   {
      FillIn editor = getActiveEditor();
      if (editor !=null)
      {
         return editor;
      }

      return this;
   }
                                         
   /**                                             
    * Reports if there is at least 1 focusable column that is not the
    * current column.
    *
    * @param    next
    *           <code>true</code> if this must search for a next focusable
    *           widget, <code>false</code> for a previous focusable widget.
    * @param    exclude
    *           This is ignored since the current column is internally known
    *           by the browse.
    *
    * @return   <code>true</code> if there is at least one other column
    *           that can be focused.
    */
   public boolean hasFocusable(boolean next, Widget exclude)
   {
      // quick exit if no edits are possible
      if (!editPossible)
         return false;
      
      // when editing is not active, browse acts like a single widget that
      // has no focusable widgets inside
      if (getActiveEditor() == null)
         return false;
      
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      // can't previous past beginning of first row
      if (!next && currRow == 0)
      {
         for (int j = currColumn - 1; j >= 0; j--)
         {
            if (!cols[j].isReadOnly())
               return true;
         }
         
         return false;
      }
      
      // can't next past end of last row
      if (next && currRow == (dcRowCount - 1))
      {
         for (int k = currColumn + 1; k < cols.length; k++)
         {
            if (!cols[k].isReadOnly())
               return true;
         }
         
         return false;
      }
      
      // since we are editable, there must be another field that can be
      // accessed
      return true;
   }
   
   /**
    * Reports if there is at least 1 focusable component that is not the
    * given component.
    *
    * @param    exclude
    *           A component to ignore or <code>null</code> if any focusable
    *           component will do.
    *
    * @return   <code>true</code> if there is at least one other component
    *           that can be focused.
    */
   @Override
   public boolean hasFocusable(Widget exclude)
   {
      return hasFocusable(true, exclude) || parent().hasFocusable(exclude); 
   }
       
   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the next widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#nextFocus()
    */
   @Override
   public void nextFocus()
   {
      if (editPossible && getActiveEditor() != null)
      {
         goDown(1);
         repaint();
      }
      else
         parent().nextFocus();
   }

   /**
    * This method is called from the in-place editor when it need to initiate
    * switching to the previous widget.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#prevFocus()
    */
   @Override
   public void prevFocus()
   {
      if (editPossible && getActiveEditor() == null)
         switchToPrevCell();
      else
         parent().prevFocus();
   }
   
   /**
    * Returns selection mode state.
    * 
    * @return Is browse in the multiple selection mode.
    */
   public boolean isSelectionMultiple()
   {
      return (selectedRowIndices != null);
   }

   /**
    * Switch multiple selection on and off.
    * 
    * @param   multiple
    *          <code>true</code> enable multiple selection,
    *          <code>false</code> - disable multiple selection.
    */
   public void setSelectionMultiple(boolean multiple)
   {
      if (browseModel == null)
         return;
      
      if (multiple && selectedRowIndices == null)
      {
         selectedRowIndices = new TreeSet<Integer>();
         if (currRow >= 0)
         {
            selectedRowIndices.add(new Integer(currRow));
         }
         
         return;
      }
      
      if (!multiple && selectedRowIndices != null)
      {
         selectedRowIndices = null;
         return;
      }
      
      // do nothing, other cases should not have any effect
      return;
   }

   /**
    * Returns is given row selected.
    * 
    * @param   rowIndex
    *          Row's index in browse model.
    * 
    * @return  Is this row selected.
    */
   public boolean isRowSelected(int rowIndex)
   {
      if (selectedRowIndices == null)
      {
         return Math.max(currRow, 0) == rowIndex;
      }

      return selectedRowIndices.contains(new Integer(rowIndex));
   }

   /**
    * Returns the count of selected rows.
    * 
    * @return  The count of selected rows.
    */
   public int getSelectedRowsCount()
   {
      if (selectedRowIndices == null)
      {
         return (dcRowCount > 0) ? 1 : 0;
      }

      return selectedRowIndices.size();
   }

   /**
    * Index in browse model for the given selected row.
    * 
    * @param   selIndex
    *          Index in sorted set of selected rows indexes.
    * 
    * @return  Browse model row index.
    * 
    * @throws  IndexOutOfBoundsException
    *          if <code>selIndex</code> is less than 0 or greater than or
    *          equal to the number of selected rows.
    */
   public int getSelectedRowIndex(int selIndex)
   {
      if (selIndex < 0 || selIndex >= getSelectedRowsCount())
      {
         throw new IndexOutOfBoundsException
            ("Invalid selected row index requested: " + selIndex + ". Only " +
             getSelectedRowsCount() + " rows are present.");
      }

      if (selectedRowIndices == null)
      {
         return currRow;
      }

      Iterator<Integer> itRow = selectedRowIndices.iterator();
      for (int i = selIndex; i > 0; --i)
      {
         itRow.next();
      }

      return itRow.next().intValue();
   }

   /**
    * Fires the DefaultAction event.
    */
   public void fireDefaultAction()
   {
      EventManager.postEvent(new KeyInput(this, 
                                    EventType.KEY_PRESSED, Keyboard.SE_DEFAULT_ACTION));
   }

   /**
    * Fires the ValueChanged event.
    */
   public void fireValueChanged()
   {
      config.setModified(true);

      ThinClient.setValueChangedEvent(new KeyInput(this, 
                                                   EventType.KEY_PRESSED, 
                                                   Keyboard.SE_VALUE_CHANGED, 
                                                   true));
   }
   
   /**
    * Return status of the updates of the current row.
    * 
    * @return  <code>true</code> if current row is modified and
    *          <code>false</code> otherwise.
    */
   public boolean isCurrentRowModified()
   {
      return (currRowUpdates != null);
   }

   /**
    * Set number of rows to display data.
    * 
    * @param   down
    *          Number of data rows.
    */
   public void setDown(int down)
   {
      if (browseModel == null)
         return;
      
      // account for the title
      if (config.isBox())
         down += 2;

      // account for labels
      if (drawLabels)
         down += ilLabelsHeight + 1;
      
      setSize(width(), down);
   }

   /**
    * Return <code>true</code> if row which has focus is selected.
    * 
    * @return  <code>true</code> if row which has focus is selected and
    *          <code>false</code> otherwise.
    */
   public boolean isFocusedRowSelected()
   {
      return isRowSelected(currRow);
   }
   
   /**
    * Return number of potential data rows in the <code>Browse</code>.
    * Amount of available data can be less that number of rows which
    * potentially can be displayed by the <code>Browse</code>.
    * 
    * @return  Number of potential rows.
    */
   public int getDown()
   {
      return ilPageRowsCount;
   }

   /**
    * Set current column to specified one.
    * 
    * @param   currentColumn
    *          Reference to <code>BrowseColumn</code> instance of specified
    *          column.
    */
   public void setCurrentColumn(BrowseColumnConfig currentColumn)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      for (int i = 0; i < cols.length; i++)
      {
         if (currentColumn == cols[i])
         {
            FillIn editor = getActiveEditor();
            
            if (editor != null)
               stopEdit();
            
            currColumn = i;
            
            if (editor != null)
               startEdit();
         }
      }
   }

   /**
    * Get currently used <code>BrowseModel</code>.
    * 
    * @return  Reference to <code>BrowseModel</code> used by
    *        <code>Browse</code>.
    */
   public BrowseModel getModel()
   {
      return browseModel;
   }

   /**
    * Return reference to previous column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getPrevColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i + 1) < cols.length)
               return cols[i + 1];
            return null;
         }
      }

      return null;
   }

   /**
    * Return reference to next column relative to given column.
    * 
    * @param   column
    *          Base column.
    * 
    * @return  Reference to next column relative to given.
    */
   public BrowseColumnConfig getNextColumn(BrowseColumnConfig column)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      for (int i = 0; i < cols.length; i++)
      {
         // found it
         if (column == cols[i])
         {
            if ((i - 1) > 0)
               return cols[i - 1];
            return null;
         }
      }

      return null;
   }
   
   /**
    * Return <code>BrowseColumn</code> for specified 0-based index.
    * 
    * @param   columnId
    *          Column index.
    * 
    * @return  <code>BrowseColumn</code> instance or <code>null</code> in
    *          case of invalid index.
    */
   BrowseColumnConfig getColumn(int columnId)
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();
      
      if (columnId >= 0 && columnId < cols.length)
         return cols[columnId];
      
      return null;
   }
   
   /**
    * Returns the active in-place editor if edit is in progress. 
    *
    * @return Reference to editor or <code>null</code>.
    */
   protected FillIn getActiveEditor()
   {
      return (FillIn) get(0);
   }

   /**
    * Trigger row display events, if any row requires it.
    * 
    * @param   oldColors
    *          The previous configured column colors.
    */
   protected void triggerRowDisplayEvents(Color[] oldColors)
   {
      // avoid recursion
      if (rowDisplayTriggered)
         return;

      ThinClient client = ThinClient.getInstance();
      BrowseColumnConfig[] columns = browseModel.getColumns();
      rowDisplayTriggered = true;
      int oldRow = currRow;

      // execute any ROW-DISPLAY triggers
      try
      {
         int offset = scrollRow - dcStartRow;
         int maxVal = Math.min(offset + ilPageRowsCount, rowDisplayed.length);
         for (int i = offset; i <  maxVal; i++)
         {
            if (rowDisplayed[i])
               continue;

            rowDisplayed[i] = true;

            int rowIdx = i + dcStartRow;

            browseModel.setCurrentRow(rowIdx);

            // restore the colors to their default values, so the trigger
            // may overwrite them
            for (int k = 0; k < oldColors.length; k++)
            {
               int cid = columns[k].getID();
               BrowseColumn bcol = (BrowseColumn) client.getWidget(cid);
               bcol.config().setDcolor(oldColors[k]);
            }

            client.invokeRowDisplayTrigger(this.config().getID());
            
            // set the colors, as they might have been modified by the 
            // trigger
            for (int k = 0; k < columns.length; k++)
            {
               int cid = columns[k].getID();
               BrowseColumn bcol = (BrowseColumn) client.getWidget(cid);
               colors[i][k] = bcol.config().getDcolor();
            }
         }
      }
      finally
      {
         // ensure cleanup is always possible
         browseModel.setCurrentRow(oldRow);
      }
   }

   /**
    * Setups scrolling to fully (as possible) show the given cell.
    * 
    * @param   cellRow
    *          The cell's row index in model.
    * @param   cellColumn
    *          The cell's column index in model.
    */
   protected void cellToView(int cellRow, int cellColumn)
   {
      int newScrollColumn = scrollColumn;
      
      if (cellColumn == 0)
      {
         newScrollColumn = 0;
      }
      else if (cellColumn != newScrollColumn)
      {
         BrowseColumnConfig[] cols = browseModel.getColumns();
         
         if (cellColumn > newScrollColumn)
         {
            newScrollColumn = cellColumn;
         }
         else
         {
            int usedX = cols[newScrollColumn].getWidthChars();
            for (int i = cellColumn; i < newScrollColumn; ++i)
            {
               usedX += cols[i].getWidthChars() + 1;
            }

            while (usedX > ilPageWidth)
            {
               usedX -= cols[newScrollColumn].getWidthChars() + 1;
               --newScrollColumn;
               if (newScrollColumn == cellColumn)
               {
                  // new scroll column may be wider than all browse
                  break;
               }
            }
         }

         if (newScrollColumn < ilFirstScrollCol)
         {
            newScrollColumn = 0;
         }
      }
      
      if (newScrollColumn != scrollColumn)
      {
         scrollColumn = newScrollColumn;
         rcLabels = null;
      }

      int newScrollRow = scrollRow;
      
      if (cellRow < scrollRow)
      {
         newScrollRow = Math.max(cellRow, 0);
      }
      else if (cellRow >= scrollRow + ilPageRowsCount)
      {
         newScrollRow = cellRow - ilPageRowsCount + 1;
      }
      
      prefetch(newScrollRow, false);
   }

   /**
    * Creates an editor for the current cell and adds it to the browse.
    * Adjusts selection data if necessary.
    */
   private void startEdit()
   {
      cellToView(currRow, currColumn);
      
      BrowseColumnConfig[] cols = browseModel.getColumns();

      if (cols[currColumn].isReadOnly())
         return;

      FillInConfig cfg = new FillInConfig(ComponentConfig.FILL_IN, null);

      if (cols[currColumn].getFormat() != null)
         cfg.setFormat(cols[currColumn].getFormat());

      cfg.setDataType(cols[currColumn].getDataType());
      cfg.setPfcolor(config.getPfcolor());
      cfg.setDcolor(config.getDcolor());
      cfg.setAutoReturn(cols[currColumn].isAutoReturn());
      cfg.setAutoZap(cols[currColumn].isAutoZap());
      cfg.setHelp(cols[currColumn].getHelp());

      int width = Math.min(cols[currColumn].getWidthChars(), ilPageWidth);
      cfg.setWidthChars(width);
      cfg.setAutoResize(false);
      
      int x = 2;
      int y = Math.max(currRow, 0) - scrollRow;
      
      if (config.isBox())
      {
         ++x;
         ++y;
      }

      if (drawLabels)
      {
         y += ilLabelsHeight + 1;
      }
      
      if (scrollColumn == 0)
      {
         for (int i = 0; i < currColumn; ++i)
         {
            x += cols[i].getWidthChars() + 1;
         }
      }
      else
      {
         x += ilPageWidth - width;
         for (int i = scrollColumn; i > currColumn; --i)
         {
            x -= cols[i].getWidthChars() + 1;
         }
      }

      cfg.setColumn(x);
      cfg.setRow(y);
      cfg.setVisible(false);
      
      FillIn editor = screen().getFactory().create(FillIn.class, cfg);

      // reset underline attribute and force Browse colors
      editor.setBrowse(this);
      editor.setValue(getValue(currColumn));
      editor.setUnderline(false);
      add(editor);

      editor.setEnabled(true);
      editor.setVisible(true);
      editor.activate();
      
      // the following call is an internal focus change and is safe
      setFocus(editor);
      
      EventManager.postEvent(new FocusEvent(editor, EventType.FOCUS_GAINED));
      columnHelp  = cols[currColumn].getHelp();

      if (columnHelp != null)
      {
         ThinClient.setHelp(columnHelp);
      }

      if (selectedRowIndices != null)
      {
         selectedRowIndices.clear();
         if (currRow >= 0)
         {
            selectedRowIndices.add(new Integer(currRow));
         }
      }
   }
   
   /**
    * Finishes edit and destroys an editor.
    * 
    * @return  <code>true</code> if editing was stopped;  <code>false</code>
    *          if we were not in edit mode.
    */
   private boolean stopEdit()
   {
      FillIn editor = getActiveEditor();
      
      if (editor == null || !editor.update())
         return false;

      BaseDataType var = editor.getValue();
      updateColumn(currColumn, var);

      remove(editor);
      
      Window.setInsertMode(Window.INSERT_MODE_OVR);
      
      return true;
   }

   /**
    * Update specified cell in current row.
    * 
    * @param   index
    *          0-based index of the cell to update.
    * @param   var
    *          New value of the cell.
    * 
    * @return  <code>true</code> if column value was changed.
    */
   private boolean updateColumn(int index, BaseDataType var)
   {
      boolean res = true;
      
      if (dcGetRow(currRow)[index].equals(var))
      { 
         // no real edit
         
         res = false;
         
         if (currRowUpdates != null)
            currRowUpdates[index] = null;
      }
      else
      {
         if (currRowUpdates == null)
            currRowUpdates = new BaseDataType[browseModel.getColumns().length];

         currRowUpdates[index] = var;
      }
      
      return res;
   }
   
   /**
    * Switches in-line editor to the previous editable cell. Initiates commit
    * of the current row if necessary.
    */
   void switchToPrevCell()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      int prevColumn;
      for (prevColumn = currColumn - 1; prevColumn >= 0; 
           --prevColumn)
      {
         if (!cols[prevColumn].isReadOnly())
            break;
      }

      do
      {
         if (prevColumn >= 0)
         {
            stopEdit();
            currColumn = prevColumn;
            startEdit();
            break;
         }
         
         // releaseCurrentRow() saves and checks a current row so
         // logical calculation order is critical here
         if (!releaseCurrentRow() || currRow == 0)
         {
            startEdit();
            break;
         }
         
         setCurrentRow(currRow - 1);
         currColumn = cols.length;
         
         while (cols[--currColumn].isReadOnly())
         {
            // empty loop body
         }
         
         startEdit();
         fireValueChanged();
      }
      while (false);
   
      repaint();
   }
   
   /**
    * Switches in-line editor to the next editable cell. Initiates commit of
    * the current row if necessary.
    */
   void switchToNextCell()
   {
      BrowseColumnConfig[] cols = browseModel.getColumns();

      int nextColumn;
      for (nextColumn = currColumn + 1; nextColumn < cols.length; ++nextColumn)
      {
         if (!cols[nextColumn].isReadOnly())
            break;
      }

      do
      {
         if (nextColumn < cols.length)
         {
            stopEdit();
            currColumn = nextColumn;
            startEdit();
            break;
         }
   
         // releaseCurrentRow() saves and checks a current row so
         // logical calculation order is critical here
         if (!releaseCurrentRow() || currRow == dcRowCount - 1)
         {
            startEdit();
            break;
         }
   
         setCurrentRow(currRow + 1);
         currColumn = 0;
         
         while (cols[currColumn].isReadOnly())
            ++currColumn;
   
         startEdit();
         fireValueChanged();
         
      } while (false);

      repaint();
   }

   /**
    * Saves the current row into data model.
    * 
    * @return  Is data checking and saving was successful.
    */
   private boolean saveCurrentRow()
   {
      if (currRowUpdates == null)
      {
         return true;
      }

      int len = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         if (currRowUpdates[i] != null)
         {
            ++len;
         }
      }

      if (len == 0)
      {
         return true;
      }

      int[] indices = new int[len];
      BaseDataType[] values = new BaseDataType[len];
      
      int p = 0;
      for (int i = 0; i < currRowUpdates.length; ++i)
      {
         BaseDataType update = currRowUpdates[i];
         if (update != null)
         {
            indices[p] = i;
            values[p] = update;
            ++p;
         }
      }

      try
      {
         dcUpdateRow(currRow, indices, values);
      }
      catch (DataValidationException ex)
      {
         // Enable internal, row-level undo.
         awaitingEndError = true;
         StringBuilder buf = new StringBuilder("Press ");
         buf.append(Keyboard.kbLabel(Keyboard.END_ERROR));
         buf.append(" to undo changes to the current row in BROWSE. (5133)");
         // TODO:  embed browse name in message text after 'BROWSE';
         // currently, config.getName() returns null.
         //buf.append(" to undo changes to the current row in BROWSE ");
         //buf.append(config.getName());
         //buf.append(". (5133)");
         ThinClient.getInstance().message(buf.toString());
         return false;
      }

      currRowUpdates = null;
      return true;
   }

   /**
    * Finishes edit and commits row to model if some updates are available.
    * Restores editor if commit has been failed.
    * 
    * @return  <code>true</code> if all actions were successful.
    */
   private boolean releaseCurrentRow()
   {
      FillIn editor = getActiveEditor();
      
      if (editor == null)
         return true;

      boolean status = stopEdit();
      
      if (!status)
         return false;
      
      return saveCurrentRow();
   }

   /**
    * Returns the data row with the given index. The row must already be in
    * the cache. 
    *
    * @param  rowIndex
    *         Row's index in model.
    *
    * @return The data row from the cache.
    */
   protected BaseDataType[] dcGetRow(int rowIndex)
   {
      return dcData[dcGetRowIndex(rowIndex)];
   }

   /**
    * Updates row in the cache and the model. Row must already be in cache.
    *
    * @param  rowIndex
    *         Row's index in model.
    * @param  cellIndices
    *         Indexes of cells to update.
    * @param  cellValues
    *         New values of cells to update.
    *
    * @throws DataValidationException
    *         New values don't passed database validation rules.
    */
   private void dcUpdateRow(int            rowIndex,
                            int[]          cellIndices,
                            BaseDataType[] cellValues)
   throws DataValidationException
   {
      BaseDataType[] row = dcGetRow(rowIndex); // index checking
      browseModel.updateRow(rowIndex, cellIndices, cellValues);

      for (int i = 0; i < cellIndices.length; ++i)
      {
         row[cellIndices[i]] = cellValues[i];
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Reference to keyboard event.
    */
   @Override
   public void onKeyPressed(KeyInput event)
   {
      int key = event.actionCode();
      
      FillIn editor = getActiveEditor();
      
      if (key == Keyboard.KA_TAB)
      {
         event.consume();
         if (releaseCurrentRow())
         {
            if (editor != null)
               startEdit(); // editor must be preserved in edit mode
            
            // the following request should be conditional
            ThinClient.nextFocus(Browse.this);
         }
         return;
      }
      
      if (key == Keyboard.KA_BACK_TAB && editor == null)
      {
         event.consume();
         if (releaseCurrentRow())
         {
            //TODO: editor can only be null at this point (see 'if' 5 lines
            // above. Requires deeper investigation.
            //if (editor != null)
            //   startEdit(); // editor must be present in edit mode
            
            // the following request should be conditional
            ThinClient.previousFocus(Browse.this);
         }
         return;
      }
      
      if (dcRowCount == 0)
      { 
         // all other keys requires the presence of current row
         return;
      }
      
      boolean invalid = true;
      
      if (editor == null)
      {
         do
         {
            if (key == Keyboard.KA_CURSOR_LEFT)
            {
               if (scrollColumn == 0)
               {
                  invalid = false;
                  break;
               }
               
               --scrollColumn;
               if (scrollColumn < ilFirstScrollCol)
               {
                  scrollColumn = 0;
               }
               
               rcLabels = null;
               break;
            }
            
            if (key == Keyboard.KA_CURSOR_RIGHT)
            {
               int colsCount = browseModel.getColumns().length - 1;
               if (scrollColumn >= colsCount)
               {
                  invalid = false;
                  break;
               }
               
               if (scrollColumn == 0)
               {
                  if (ilFirstScrollCol > colsCount)
                  {
                     invalid = false;
                     break;
                  }
                  scrollColumn = ilFirstScrollCol;
               }
               else
               {
                  ++scrollColumn;
               }
               
               rcLabels = null;
               break;
            }
            
            if (key == Keyboard.KA_CURSOR_UP)
            {
               if (currRow <= 0)
               {
                  invalid = false;
                  break;
               }
               
               setCurrentRow(currRow - 1);
               cellToView(currRow, scrollColumn);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_CURSOR_DOWN)
            {
               if (currRow >= dcRowCount - 1)
               {
                  invalid = false;
                  break;
               }
               
               setCurrentRow(currRow + 1);
               cellToView(currRow, scrollColumn);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_HOME)
            {
               if (currRow == 0)
               {
                  invalid = false;
                  break;
               }
               
               setCurrentRow(0);
               cellToView(0, scrollColumn);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_END)
            {
               setCurrentRow(Integer.MIN_VALUE);
               cellToView(currRow, scrollColumn);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_RIGHT_END)
            {
               setCurrentRow(-ilPageRowsCount);
               cellToView(currRow, scrollColumn);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_PAGE_UP)
            {
               if (currRow <= 0)
               {
                  invalid = false;
                  break;
               }

               resetRowDisplayState();

               int oldScroll = scrollRow;
               scrollRow -= ilPageRowsCount;
               
               if (scrollRow < 0) 
                  scrollRow = 0;
               
               // PAGE-UP/DOWN special case - set flags here
               rowDisplayTriggered = scrollRow == oldScroll;
               ignoreQuirk = true;

               cellToView(scrollRow, scrollColumn);
               setCurrentRow(currRow - ilPageRowsCount);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_PAGE_DOWN)
            {
               int rowCount = dcRowCount - 1;
               if (currRow >= rowCount)
               {
                  invalid = false;
                  break;
               }

               resetRowDisplayState();

               if (currRow > rowCount)
               {
                  setCurrentRow(rowCount);
               }
               
               int oldScroll = scrollRow;
               scrollRow += ilPageRowsCount;
               
               if (scrollRow > rowCount)
               { // mimics Progress: whole page or nothing
                  scrollRow -= ilPageRowsCount;
               }
               
               // PAGE-UP/DOWN special case - set flags here
               rowDisplayTriggered = scrollRow == oldScroll;
               ignoreQuirk = true;

               cellToView(scrollRow, scrollColumn);
               setCurrentRow(currRow + ilPageRowsCount);
               
               if (!isSelectionMultiple())
                  fireValueChanged();
               
               break;
            }
            
            if (key == Keyboard.KA_REPLACE)
            {
               if (!editPossible)
               {
                  invalid = false;
                  break;
               }
               
               config.setReadOnly(false);
               break;
            }
            
            if (key == Keyboard.SE_ENTRY)
            {
               fireDefaultAction();
               break;
            }
            
            return; // ignoring other keys
         }
         while (false);
      }
      else
      {
         invalid = false;
         
         do
         {
            if (key == Keyboard.KA_BACK_TAB)
            {
               switchToPrevCell();
               break;
            }
            
            if (key == Keyboard.KA_EDITOR_TAB)
            {
               switchToNextCell(); // this includes redraw
               break;
            }
            
            if (key == Keyboard.KA_HOME)
            {
               if (!releaseCurrentRow() || currRow == 0)
                  break;
               
               stopEdit();
               cellToView(0, scrollColumn);
               setCurrentRow(0);
               startEdit();
               fireValueChanged();
               invalid = true;
               break;
            }
            
            if (key == Keyboard.KA_END)
            {
               if (!releaseCurrentRow())
                  break;
               
               stopEdit();
               setCurrentRow(Integer.MIN_VALUE);
               cellToView(currRow, scrollColumn);
               startEdit();
               fireValueChanged();
               invalid = true;
               break;
            }
            
            if (key == Keyboard.KA_RIGHT_END)
            {
               if (!releaseCurrentRow() || currRow == dcRowCount - 1)
                  break;
               
               stopEdit();
               setCurrentRow(-ilPageRowsCount);
               cellToView(currRow, scrollColumn);
               startEdit();
               fireValueChanged();
               invalid = true;
               break;
            }
            
            if (key == Keyboard.KA_CURSOR_UP)
            {
               invalid = goUp(1);
               break;
            }
            
            if (key == Keyboard.KA_PAGE_UP)
            {
               invalid = goUp(ilPageRowsCount);
               break;
            }
            
            if (key == Keyboard.KA_CURSOR_DOWN)
            {
               invalid = goDown(1);
               break;
            }
            
            if (key == Keyboard.KA_PAGE_DOWN)
            {
               invalid = goDown(ilPageRowsCount);
               break;
            }
            
            if (key == Keyboard.KA_REPLACE)
            {
               config.setReadOnly(true);
               releaseCurrentRow();
               invalid = true;
               break;
            }
            
            if (key == Keyboard.KA_END_ERROR && awaitingEndError)
            {
               awaitingEndError = false;
               undoValue();
               break;
            }
            
            return; // ignoring other keys
         }
         while (false);
      }
      
      event.consume(); // consume all handled keys
      
      if (invalid)
      {
         draw();
         requestSync();
      }
      else if (getActiveEditor() != null)
      {
         editor = getActiveEditor(); // editor may be recreated
         editor.draw();
         requestSync();
      }
   }
   
   /**
    * Move cursor up by the number of rows specified, or as many as possible,
    * if there are fewer rows available than the size of the requested move.
    * 
    * @param   size
    *          Number of rows to go up, relative to the cursor's current
    *          position.
    *          
    * @return  <code>true</code> if current position is changed.
    */
   private boolean goUp(int size)
   {
      if (size == 0)
      {
         // no move requested
         return false;
      }
      
      // implies stopEdit()
      if (!releaseCurrentRow())
      {
         startEdit();
         return false;
      }
      
      setCurrentRow(currRow - size);
      
      if (size > 1)
      {
         scrollRow -= size;
         
         if (scrollRow < 0) 
            scrollRow = 0;
         
         cellToView(scrollRow, scrollColumn);
      }
      
      cellToView(currRow, scrollColumn);
      
      stopEdit();
      startEdit();
      fireValueChanged();
      return true;
   }
   
   /**
    * Move cursor down by the number of rows specified, or as many as
    * possible, if there are fewer rows available than the size of the
    * requested move.
    * 
    * @param   size
    *          Number of rows to go down, relative to the cursor's current
    *          position.
    *          
    * @return  <code>true</code> if current position is changed.
    */
   private boolean goDown(int size)
   {
      if (size == 0)
      {
         // no move requested
         return false;
      }
      
      // implies stopEdit()
      if (!releaseCurrentRow())
      {
         startEdit();
         return false;
      }

      int rowCount = dcRowCount - 1;
      
      setCurrentRow(currRow + size);
      
      if (size > 1)
      {
         scrollRow += size;
         
         if (scrollRow > rowCount)
            // mimics Progress: whole page or nothing
            scrollRow -= size;
         
         cellToView(scrollRow, scrollColumn);
      }
      else
      {
         cellToView(currRow, scrollColumn);
      }
      
      startEdit();
      fireValueChanged();
      
      return true;
   }
   
   /**
    * Intercept special events. Using regular keyTyped/keyPressed may not work
    * if component is not in focus.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#processEvent(com.goldencode.p2j.ui.client.event.Event)
    */
   @Override
   public void processEvent(Event evt) 
   {
      if (evt instanceof KeyInput)
      {
         KeyInput ki = (KeyInput) evt;
         int action = ki.actionCode();
         
         if(action == Keyboard.SE_VALUE_CHANGED ||
            action == Keyboard.SE_ENTRY)
         {
            // if widget is not initialized then do nothing
            if (getActiveEditor() != null || browseModel == null)
            {
               return;
            }

            // ignore VALUE-CHANGED events generated by Browse, 
            // they are for external consumers, all internal actions
            // are already performed 
            if (action == Keyboard.SE_VALUE_CHANGED && ki.isSpecial())
            {
               return;
            }
            
            refreshStartRow(true);
            return;
         }
         
         onKeyPressed(ki);
         boolean consumed = ki.isConsumed();
         
         if (consumed                     || 
             action == Keyboard.KA_RETURN || 
             action == Keyboard.keyCode(" "))
         {
            // if the key was processed by the browse, clear the message area
            // if space or return was pressed, clear the message area
            ThinClient.getInstance().clearPutScreenInMessageArea();            
         }
         
         if (consumed)
            return;
      }
      
      super.processEvent(evt);
   }
   
   /**
    * Handles ' ' key in R/O mode: fires ValueChanged event. Additionally
    * inverts selection status of the current line in multiple selection
    * mode.
    *
    * @see com.goldencode.p2j.ui.client.event.listener.KeyListener#onKeyTyped(KeyInput)
    */
   @Override
   public void onKeyTyped(KeyInput event)
   {
      if (dcRowCount == 0)
      {
         return;
      }
      
      if (event.keyCode() == ' ' && getActiveEditor() == null)
      {
         if (selectedRowIndices != null)
         {
            Integer invIndex = new Integer(currRow);
            
            if (selectedRowIndices.contains(invIndex))
            {
               selectedRowIndices.remove(invIndex);
            }
            else
            {
               selectedRowIndices.add(invIndex);
            }
         }
         
         fireValueChanged();
         draw();
         requestSync();
      }
   }

   /**
    * Track all changes in the current row.
    *
    * @param   currRow
    *          New current row.
    */
   private void setCurrentRow(int currRow)
   {
      setCurrentRow(currRow, false);
   }

   /**
    * Track all changes in the current row.
    * 
    * @param   currRow
    *          New current row.
    * @param   force
    *          Call server-side row setting facility even if on the client
    *          side the new current row equals the current row. 
    */
   private void setCurrentRow(int currRow, boolean force)
   {
      if (currRow > Integer.MIN_VALUE && currRow < 0)
      {
         // Integer.MIN_VALUE has special meaning:  set last row in browse
         // as current row.  All other negative numbers must be clipped to
         // 0 to select first record.
         currRow = 0;
      }
      
      if ((this.currRow != currRow || force) &&
          dcRowCount > 0 &&
          browseModel != null)
      {
         this.currRow = browseModel.setCurrentRow(currRow);
      }
   }
   
   /**
    * Implements the <b>SET-REPOSITIONED-ROW() </b> <b>BROWSE </b> widget
    * method.
    * 
    * @param   n
    *          0-based row number where the new record is displayed
    * @param   conditional
    *          if <code>true</code>, tries to reposition to the browse
    *          viewport existing row first
    */
   public void selectRepositionedRow(int n, boolean conditional)
   {
      // TODO: implement it
   }

   /**
    * Get text value of the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * 
    * @return  The text value of the given column in current row.
    */
   public String getColumnText(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().getText(); 
         
      cellToView(currRow, ordinal);
      
      String fmt = browseModel.getColumns()[ordinal].getFormat();
      BaseDataType value = null;
      
      value = getValue(ordinal);

      if (value == null)
         return "";

      return (fmt == null) ? 
                            character.valueOf(value).getValue():
                            character.valueOf(value, fmt).getValue(); 
   }

   /**
    * Get value for the specified column.
    * 
    * @param   index
    *          The 0-based column index.
    *          
    * @return  Column value.
    */
   private BaseDataType getValue(int index)
   {
      if (currRowUpdates != null && currRowUpdates[index] != null)
         return currRowUpdates[index];
      
      return dcGetRow(currRow)[index];
   }

   /**
    * Get value of the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * 
    * @return  The value of the given column in current row.
    */
   public BaseDataType getColumnValue(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().getValue(); 
         
      cellToView(currRow, ordinal);
      
      BaseDataType value = null;
      
      value = getValue(ordinal);

      if (value == null)
         return new character();

      return value; 
   }

   /**
    * Set value for the given column in current row.
    * 
    * @param   ordinal
    *          The column ordinal.
    * @param   value         
    *          The value for the cell.
    */
   public void setColumnValue(int ordinal, BaseDataType value)
   {
      if (config.isReadOnly())
         return;
      
      if (ordinal == currColumn && getActiveEditor() != null)
      {
         getActiveEditor().setValue(value);
         return;
      }
      
      if (browseModel.getColumns()[ordinal].isReadOnly())
         return;

      updateColumn(ordinal, value);
   }

   /**
    * Check if the value in specified column in current row is changed. 
    * 
    * @param   ordinal
    *          The column ordinal.
    *          
    * @return  <code>true</code> if value is changed.
    */
   public boolean isModified(int ordinal)
   {
      if (ordinal == currColumn && getActiveEditor() != null)
         return getActiveEditor().isModified(); 
         
      cellToView(currRow, ordinal);

      if (currRowUpdates != null && currRowUpdates[ordinal] != null)
         return true;

      return false;
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    */
   public void refresh()
   {
      refresh(false, true);
   }
   
   /**
    * Refresh the browse cache and redisplay the browse.
    * 
    * @param   setStartRow
    *          If set to <code>true</code> set start row to current row in
    *          the query.
    * @param   forceReloadRowData
    *          Force reload row data. 
    */
   public void refresh(boolean setStartRow, boolean forceReloadRowData)
   {
      // you can get to this function in four ways:
      // 1. query has been repositioned
      // 2. Progress browse:refresh() method has been called
      // 3. query has been closed
      // 4. postponed refresh - means that we was in this function before but
      //    we have postponed refresh because the browse was hidden

      // ignore request if editing is active
      if (browseModel == null)
         return;

      Frame frame = (Frame) UiUtils.locateFrame(this);
      
      if (frame == null)
         return;
      
      if (!frame.isVisible())
      {
         postponedRefresh = true;
         if (setStartRow)
            postponedRowReset = true;
         return;
      }

      boolean editable = getActiveEditor() != null;
      if (editable)
      {
         stopEdit();

         config.setModified(false);
         currRowUpdates = null;
      }

      if (postponedRefresh)
      {
         setStartRow = postponedRowReset;
      }
      
      postponedRefresh = false;
      postponedRowReset = false;
      
      int rows    = browseModel.getRowCount(dcStartRow);
      int current = browseModel.getCurrentRow();
      boolean needRedraw;

      if (hadDraw && dcStartRow == 0 && !pseudoFocus)
      {
         // query is initialized after first draw() call
         pseudoFocus = true;
         separatorQuirk = true;
      }
                                                            
      if (setStartRow && rows > ilPageRowsCount)
      {
         needRedraw = (dcRowCount != rows || scrollRow != current);
         needRedraw = initCache(current,
                                current,
                                needRedraw || forceReloadRowData)
                      || needRedraw;
      }
      else
      {
         needRedraw = (dcRowCount != rows);
         needRedraw = initCache(scrollRow,
                                current,
                                needRedraw || forceReloadRowData)
                      || needRedraw;
      }

      if (editable)
      {
         startEdit();
      }
      
      if (isVisible() && needRedraw)
      {
         draw();
         requestSync();
      }
   }

   /**
    * Refresh start row in the browse and redisplay.
    * 
    * @param    force
    *           Force refresh.
    */
   public void refreshStartRow(boolean force)
   {
      if (force)
         initCache(scrollRow, browseModel.getCurrentRow(), true);
      else
      {
         resetRowDisplayState();
         refresh(true, false);       
      }
      
      repaint();
   }
   
   /**
    * Reset the cache and redisplay.
    */
   public void queryClosed()
   {
      stopEdit();
      
      resetRowDisplayState();
      
      this.dcRowCount = 0;
      this.dcData = null;
      this.currRow = -1;
      this.scrollRow = 0;
      refresh(true, false);
      repaint();
   }

   /**
    * Reset the state of the rows, as when no row has been displayed.
    */
   protected void resetRowDisplayState()
   {
      if (rowDisplayed != null)
      {
         Arrays.fill(rowDisplayed, false);
      }
      oldDisplayRow = -1;
   }
   
   /**
    * Initialize row data cache.
    * 
    * @param   row
    *          New start row for the view.
    * @param   current
    *          New current row to display.
    * @param   force
    *          Force reloading of records.
    *          
    * @return  <code>true</code> if data, initial or current row were changed
    *          or cache was reloaded and redrawing is required.
    */
   private boolean initCache(int row, int current, boolean force)
   {
      //dcData = null;
      
      boolean needDraw = !(scrollRow == row && current == currRow) || force;
      
      int newScrollRow = Math.max(0, row);

      needDraw = prefetch(newScrollRow, force) && needDraw;

      if (row != current)
         cellToView(current, scrollColumn);
      
      setCurrentRow(current); 
      
      return needDraw;
   }

   /**
    * Load specified row into the cache.
    * <p>
    * General caching approach is window based and aligned on the page
    * boundary. In other words, cache always represents continuous set of rows
    * containing at most <code>CACHE_SIZE</code> pages where all except,
    * perhaps, last, pages contain ilPageRowsCount rows. Last page may contain
    * less rows. When requested row belongs to page which is not present in
    * cache, then entire cache window is shifted to the new location.
    * 
    * @param   row
    *          Index of the row in the data model.
    * @param   force
    *          Force full reload.
    * 
    * @return  <code>true</code> if data was changed or reloaded and
    *          redrawing is required.
    * 
    */
   private boolean prefetch(int row, boolean force)
   {
      force = force || (row != scrollRow);
      
      // if this is the initial draw or the scroll row changed, force
      // ROW-DISPLAY trigger execution; do this only if the quirk for the
      // PAGE-UP/DOWN is not ON
      if (!ignoreQuirk)
      {
         rowDisplayTriggered = hadDraw && row == scrollRow;
      }
      ignoreQuirk= false;
      scrollRow = row;
      
      int start = row - row % ilPageRowsCount;
      int pageEnd = Math.min(dcRowCount, start + ilPageRowsCount);
      
      // check if requested row is in cache
      if (!force &&
          dcData != null &&
          start >= dcStartRow &&
          pageEnd <= (dcStartRow + dcData.length))
      {
         // Note that due to page alignment presence of requested row
         // in cache means that entire page is also in cache.

         // no need to redraw it
         return false;
      }
      
      dcRowCount = browseModel.getRowCount(start);
      dcStartRow = start;
      
      int count = Math.max(0, dcRowCount - start);
      count = Math.min(count, ilPageRowsCount * 3);
      
      dcData = browseModel.getRows(start, count);

      BrowseColumnConfig[] columns = browseModel.getColumns();
      int colNo = columns.length;

      Color[][] oldColors = colors;
      colors = new Color[dcData.length][colNo];
      rowDisplayed = new boolean[dcData.length];

      int oldStart = oldDisplayRow;
      int oldEnd = oldStart == -1 ? -1 : oldStart + ilPageRowsCount;
      // as the query might have moved to other rows, we must  determine if a 
      // row is already displayed on the screen or not. also, their color must 
      // be preserved, so the next drawing will not corrupt its color.
      for (int i = 0; i < rowDisplayed.length; i++)
      {
         int rowNo = i + dcStartRow;
         rowDisplayed[i] = rowNo >= oldStart && rowNo < oldEnd;

         if (rowDisplayed[i])
         {
            // preserve the color, if it was computed
            int oldRowPosition = i - (oldDisplayRow - scrollRow);
            if (oldRowPosition >= 0 && oldRowPosition < oldColors.length)
            {
               colors[i] = Arrays.copyOf(oldColors[oldRowPosition], colNo);
            }
         }
      }
      
      return true;
   }

   /**
    * Calculate row index for internal cache.
    * 
    * @param   rowIndex
    *          Index in the model.
    *          
    * @return  Calculated index in the local cache.
    */
   private int dcGetRowIndex(int rowIndex)
   {
      rowIndex = Math.max(rowIndex, 0);
      
      if (rowIndex < dcStartRow || rowIndex >= dcStartRow + dcData.length)
      {
         throw new IllegalArgumentException
            ("Invalid cache read request: " + rowIndex +
             ". Cache contains only [" + dcStartRow + ":" +
             (dcStartRow + dcData.length) + ")");
      }

      return rowIndex - dcStartRow;
   }
}
