/*
** Module   : StatsWorker.java
** Abstract : provides an interface for maintaining sets of statistics from
**            pattern engine rulesets
**
** Copyright (c) 2005-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 GES 20050225   @20124 Created initial version supporting functions
**                           for manually and automatically creating/ 
**                           matching statistics as well as reporting
**                           functions for summary and detailed reports.
** 002 GES 20050303   @20197 Second pass. Converted from file indices to
**                           embedded filenames and a major rewrite with
**                           generation of HTML output using the Apache
**                           Element Construction Set (ECS). Also added
**                           sorting options for the summary stats,
**                           parser type output in detailed records and
**                           the ability to store user-defined text as
**                           the match text.
** 003 GES 20050310   @20284 Minor change to provide control over source
**                           file cross-reference generation.
** 004 GES 20050311   @20291 Fixed matches comparator to make the equals
**                           behavior correct (was eliminating all but
**                           1 entry that had the same number of matches).
**                           Also added a check in copyToHtml() to
**                           bypass this if the target exists and is
**                           later in timestamp than the source.
** 005 GES 20050316   @20399 Converted this worker to use the visitAst()
**                           interface in PatternWorker instead of using
**                           a rule driven approach. Also added AST spec
**                           output to the footer.
** 006 GES 20050318   @20429 Made a minor change to move dump tree
**                           support into AnnotatedAst and use that
**                           common method.                            
** 007 GES 20050318   @20442 Added support to save off the condition
**                           during match processing and to read those
**                           conditions back and write them into the
**                           footer of a details report.
** 008 ECF 20050405   @20647 Added second addMatchMultiplexed method. This
**                           version accepts custom name and type strings
**                           rather than extracting the name from the
**                           current AST's token type.
** 009 ECF 20050430   @20973 Modified due to changes in pattern package.
**                           A PatternEngine reference is no longer
**                           maintained, but is retrieved from the
**                           AstSymbolResolver instead when needed.
** 010 ECF 20050526   @21290 Changes to support new expression engine
**                    @21305 implementation. Changed library registration
**                           mechanism based on superclass' modifications
**                           to support single library per pattern worker
**                           limit. This involved merging the Stats and
**                           Report inner classes.
** 011 NVS 20050531   @21333 addMatchMultiplexed method which takes a cus-
**                           tom name now translates the name so it can't
**                           contain slashes. Slashes would produce a file
**                           name that is impossible to create. The origi-
**                           nal name is still used as a long name in the
**                           report, because it's safe.
** 012 ECF 20050722   @21820 Added statistical analysis user functions.
**                           The methods maximum, minimum, median, mean,
**                           and standardDeviation implement user
**                           functions which are backed by methods of the
**                           same names in StatisticsHelper.
** 013 ECF 20050727   @21893 Added loadDistributions and population user
**                           functions. The former loads a map of
**                           distribution data from an XML document. The
**                           latter accumulates total population size of a
**                           distribution.
** 014 GES 20060129   @24130 FileList usage changes.
** 015 GES 20070514   @33504 Added a reset statistics method to enable
**                           running multiple different reports in the
**                           same pipeline. Added support for generating
**                           a master report index. Added a way to
**                           override the current condition for a report.
**                           Only a single source file index is being
**                           emitted now (with a std name).
** 016 GES 20070626   @34287 If file_xref is active, write the source
**                           file index link into each page footer instead
**                           of the AST list. The AST list is a fallback
**                           if file_xref is off. Also fixed a minor
**                           formatting issue with the AST list text.
**                           Made a small improvement to the timestamp
**                           formatting. Added support to control the
**                           filename of the source code file cross-ref.
**                           Added mutual-exclusion flag to support the
**                           calc and display of percentages in the
**                           summary report. Fixed the URL and filenames
**                           used for detail reports when the text
**                           contained invalid characters.
** 017 GES 20070627   @34297 Custom report creation based on user defined
**                           data, but leaving the report formatting and
**                           HTML work to this worker.
** 018 GES 20070628   @34320 Made inner class static otherwise it pins
**                           all StatsWorker (and PatternEngine...) etc
**                           into memory. This is because a non-static
**                           inner class has an implicit reference to
**                           its containing class' instance (this). This
**                           greatly reduces the working set of reports
**                           created via more than 1 pattern engine run.
** 019 ECF 20070703   @34355 Fixed file resource leak. Output streams were
**                           not being closed after reports were written.
** 020 GES 20070710   @34436 Implemented a better algorithm to determine
**                           an artificial node's line/column number.
**                           Handled two other cases of file leaks (see
**                           H019).
** 021 GES 20070711   @34477 Rewrite to replace ECS with our own HTML
**                           helpers. This is needed because the ECS
**                           approach of building the entire output in
**                           memory before streaming it out to file is
**                           extremely memory intensive on large reports.
**                           This approach leads to out of memory errors
**                           so we are moving to a "stream it as you go"
**                           to eliminate this issue.
** 022 GES 20070713   @34493 Each report (summary and its associated
**                           details) are now created in a separate and
**                           unique subdirectory.
** 023 GES 20070714   @34496 Long detail reports are now paginated to
**                           ensure that they are never too big to load
**                           and use in a browser.
** 024 GES 20070716   @34529 Rework to move source file copy processing
**                           to a delegated model so that this code
**                           doesn't have to embed knowledge of each
**                           possible source file type. Instead, an object
**                           is registered to provide the AST-specific
**                           processing needed. This allows non-4GL source
**                           ASTs to be processed via the reporting code.
** 025 GES 20070720   @34645 Default page size is now 100. Added columns
**                           to the table of contents for paginated
**                           reports. These show the start and end files
**                           for each page.  Made the auto-generated
**                           multiplex text for a statistic more 
**                           descriptive.
** 026 GES 20080309   @37670 Switched to using character readers instead
**                           of byte streams. This handles I18N needs.
** 027 GES 20090429   @42069 Match package and class name changes.
** 028 GES 20090515   @42237 Moved to AstManager from AstRegistry/AstPersister.
** 029 ECF 20111206          Hastily hacked in temporary database support
**                           until we can do a better job of this.
** 030 GES 20111207          Fixed issue with generateSafeFilename() where the
**                           resulting filename was too long for the file sys.
**                           Made master index processing persist its state
**                           across multiple JVM relaunches. This allows a
**                           restart of the reporting batch process after a
**                           failure or other abend and still the master index
**                           can be generated correctly after the last report
**                           (it will contain all of the reports done in the
**                           batch even if there were restarts).
*/

package com.goldencode.p2j.util;

import java.io.*;
import java.sql.*;
import java.util.*;
import java.util.Date;
import java.text.*;
import com.goldencode.ast.*;
import com.goldencode.html.*;
import com.goldencode.p2j.io.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.cfg.*;

/**
 * Provides users of the <code>PatternEngine</code> with a set of
 * user-defined functions and variables for manually and/or automatically 
 * creating/matching statistics as well as reporting functions for both
 * summary and detailed reports.
 * <p>
 * See {@link Match} and {@link Statistic} which are core container and
 * helper classes providing basic data services to this worker.
 */
public class StatsWorker
extends AbstractPatternWorker
{
   /** Cache file for master index and categories data. */
   public static final String MASTER_INDEX_CACHE = "master_index.cache";
   
   /** Limit to the filename portion of a generated HTML path. */
   private static final int FILENAME_LIMIT = 128;
   
   /** Stores summary report data to build a master index of reports. */
   private static List masterIndex = null;
   
   /** Stores all categories used in the master index of reports. */
   private static Map categories = null;
   
   /** Default database user. */
   private static String dbUser = "sa";
   
   /** Default database password. */
   private static String dbPass = "";
   
   /** Database URL */
   private String dbURL = null;
   
   /** Anchor name associated with the all reports case. */
   private String allReportsAnchor = null;
      
   /** Stores statistics based on a key of a short name. */
   private Map stats = null;
   
   /** Quick access to the registry for AST ID and filename lookups. */
   private AstManager mgr = null;
   
   /**
    * Stores a list of all files processed, even if no matches were found
    * in a given file.
    */
   private Set filelist = null;
   
   /** A given report run needs a helper to create the source file index. */
   private SourceFileHelper srcHelper = null;
   
   /** Headings for the source file index columns. */
   private String[] headings = null;
   
   /** Next filename ID */
   private int nextFileID = 0;
   
   /** Filename to ID mappings */
   private Map<String, Integer> file2ID = new HashMap<String, Integer>();
   
   /** ID to filename mappings */
   private Map<Integer, String> id2File = new HashMap<Integer, String>();
   
   /** Database connection */
   private Connection connection = null;
   
   /** Database insert statement */
   private PreparedStatement stmtInsert = null;
   
   /** Database query statement for records by statistic */
   private PreparedStatement stmtQueryByStat = null;
   
   /** Database query statement for records by file (within a statistic) */
   private PreparedStatement stmtQueryByFile = null;
   
   /** Flag indicating whether database table has been indexed/analyzed */
   private boolean indexed = false;
   
   /**
    * Default constructor which calls the super-class constructor, registers
    * its libraries and initializes its instance members.
    */
   public StatsWorker()
   throws AstException
   {
      super();
      
      Library helper = new Library();
      setLibrary(helper);
      
      stats    = new LinkedHashMap();
      mgr      = AstManager.get();      
      filelist = new LinkedHashSet();
      
      prepareDatabase();
      
      // this constructor is executed with each new report, this is the
      // optimal time to save off any previous master index results so that
      // if this report fails, we can restart without losing anything
      if (masterIndex != null)
      {
         try
         {
            helper.saveMappings(MASTER_INDEX_CACHE);
         }
         
         catch (Exception exc)
         {
            exc.printStackTrace(System.err);
         }
      }
   }
   
   protected void finalize()
   {
      closeConnection();
   }
   
   /**
    * Provides a hook for notification of the worker when a new AST is
    * being processed.  This is used to maintain a list of all files
    * that have been processed, regardless of whether any matches are
    * found within the file or not.
    *
    * @param   ast
    *          The root node of the source AST about to be processed by the
    *          pattern engine.
    */
   public void visitAst(Aast ast)
   {
      String filename = ast.getFilename();
      
      if (filename == null)
      {
         // TODO: log this
         System.err.println("AstRegistry failure for ID " +
                            ast.getId().toString() + "!");
      }
      else
      {
         filelist.add(filename);
      }
   }
   
   /**
    * Opens a connection to the H2 database (in the ./rptdb/rptdb file) and
    * initializes all needed prepared statements.
    * 
    * @throws  RuntimeException if there were errors during db preparation. 
    */
   private void prepareDatabase()
   {
      long maxMemory = Runtime.getRuntime().maxMemory();
      long cacheSize = -1;
      if (maxMemory < Long.MAX_VALUE)
      {
         cacheSize = maxMemory / 1024 / 4;
      }
      
      dbURL =
         "jdbc:h2:rptdb/rptdb" +
         ";DB_CLOSE_DELAY=-1;AUTOCOMMIT=OFF;LOG=0;LOCK_MODE=0;UNDO_LOG=0";
      
      if (cacheSize > -1)
      {
         dbURL += (";CACHE_SIZE=" + cacheSize);
      }
      
      Statement stmt = null;
      
      try
      {
         // load driver
         Class.forName("org.h2.Driver");
         
         // this connection stays open for the life of the report run;  it is
         // closed when this object is gc'd (yes, I know, ugly, but we don't
         // have a better hook)
         openConnection();
         stmt = connection.createStatement();
         
         // create table
         String create =
            "create cached local temporary table match (" +
            "id identity, stat int, file int, line int, col int, text varchar)";
         stmt.execute(create);
         
         // prepare insert statement
         String insert =
            "insert into match (stat, file, line, col, text) " +
            "values(?, ?, ?, ?, ?)";
         stmtInsert = connection.prepareStatement(insert);
         
         // prepare query statements
         String query =
            "select file, line, col, text from match " +
            "where stat = ? order by file asc, line asc, col asc";
         stmtQueryByStat = connection.prepareStatement(query);
         query =
            "select line, col, text from match " +
            "where stat = ? and file = ? order by line asc, col asc";
         stmtQueryByFile = connection.prepareStatement(query);
         
         connection.commit();
      }
      catch (Exception exc)
      {
         try
         {
            connection.rollback();
         }
         catch (SQLException sqle)
         {
            // ignore
         }
         
         throw new RuntimeException("Error preparing database", exc);
      }
      finally
      {
         if (stmt != null)
         {
            try
            {
               stmt.close();
            }
            catch (SQLException exc)
            {
               // ignore
            }
         }
      }
   }
   
   /**
    * Open a new connection to the report database (if one already opened,
    * it will close it first).
    * 
    * @throws  RuntimeException if new connection could not be established. 
    */
   private void openConnection()
   {
      if (connection != null)
      {
         closeConnection();
      }
      
      try
      {
         connection = DriverManager.getConnection(dbURL, dbUser, dbPass);
      }
      catch (Exception exc)
      {
         throw new RuntimeException("Error opening database connection", exc);
      }
   }
   
   /**
    * Close the current connection
    *  
    * @throws  RuntimeException if connection could not be closed. 
    */
   private void closeConnection()
   {
      if (connection == null)
      {
         return;
      }
      
      try
      {
         connection.close();
      }
      catch (Exception exc)
      {
         throw new RuntimeException("Error closing database connection", exc);
      }
   }
   
   /**
    * Creates an index on the stat and file columns for the match table,
    * and runs analyze on the database.
    * 
    * @throws  RuntimeException if there were errors during index creation or
    *          analyze. 
    */
   private void indexAndAnalyze()
   {
      if (indexed)
      {
         return;
      }
      
      indexed = true;
      
      Statement stmt = null;
      
      try
      {
         // create index
         stmt = connection.createStatement();
         String index =
            "create index idx_match_1 on match (stat asc, file asc)";
         stmt.execute(index);
         
         // analyze (also commits)
         stmt.execute("analyze");
      }
      catch (Exception exc)
      {
         try
         {
            connection.rollback();
         }
         catch (SQLException sqle)
         {
            // ignore
         }
         
         throw new RuntimeException("Error indexing and analyzing", exc);
      }
   }
   
   /**
    * Provides a service for defining and matching statistics on a manual 
    * basis (see {@link #addStatistic} and {@link #addMatch}) and/or on
    * an automated basis (see {@link #addMatchMultiplexed}). Provides helpers
    * to generate statistics reports. This class is a
    * <code>PatternEngine</code> worker library.
    */
   public class Library
   implements HtmlConstants
   {
      /** Normal text processing in fields. */
      private final int NORMAL    = 0;
      
      /** Bold text processing in fields. */
      private final int BOLD      = 1;
      
      /** Uses a preformatted element for text processing in fields. */
      private final int PREFORMAT = 2;
      
      /** Sort descending using aggregate matches. */
      private final int SORT_MATCHES = 0;
      
      /**
       * Sort lexicographically (in alphanumeric order) by statistic short
       * name.
       */
      private final int SORT_ALPHA = 1;
      
      /** Sort in the natural insertion order of the statistics. */
      private final int SORT_INSERTION = 2;
      
      /** Olive color specification. */
      private final String OLIVE = "#999900";
      
      /** Gold color specification. */
      private final String GOLD = "#FFCC66";
      
      /** White color specification. */
      private final String WHITE = "#FFFFFF";
      
      /** Default source file index filename. */
      private final String INDEX_FILENAME = "source_file_index.html";
      
      /** Heading for the list of all possible reports. */
      private final String ALL_RPTS = "All Reports";
      
      /** Standard summary base file name. */
      private final String SUM_NAME = "index.html";
      
      /** Standard prefix for each detail report base file name. */
      private final String DET_PREFIX = "detail_";
      
      /** Date format string. */
      private final String FMT = "MM/dd/yyyy HH:mm:ss z";
      
      /** Date and time formatter. */
      private final SimpleDateFormat sdf = new SimpleDateFormat(FMT);
      
      /** Formats a number padded on left with zeros. */
      private NumberFormat fmt = null;      
      
      /** Default output directory or <code>null</code>. */
      private String dir = null;
      
      /** Comparator for lexicographic sorting. */
      private Comparator alphaCompare = null;
      
      /** Comparator for sorting in descending order by aggregate matches. */
      private Comparator matchCompare = null;
      
      /** Type of sorting that should be used. */
      private int sortOrder = SORT_MATCHES;
      
      /** 
       * Determines if a master source file cross-reference will be
       * automatically generated.
       */
      private boolean file_xref = false;
      
      /** Filename for the source file cross-reference. */ 
      private String xref_name = INDEX_FILENAME;
      
      /** Global override for the condition expression text. */
      private String conditionText = null;
      
      /** Mutual-exclusion flag to enable/disable percent calculations. */
      private boolean mutex = false;
      
      /** All categories associated with the currently running report. */
      private Set current_cats = null;
      
      /** Maximum number of matches per page in a detail report. */
      private int pageSize = 100;
      
      /** Index to differentiate truncated detailed report filenames. */
      private int counter = 0;
      
      /**
       * Constructs a default instance.  Initializes formatting members to
       * known values.
       * <p>
       * This method initializes HTML entity conversion to ensure that all
       * characters can be properly rendered.  The following subset are
       * converted at this time (more can be added as needed):
       * <p>
       * &amp;
       * &quot;
       * &lt;
       * &gt;
       * &apos;
       */
      public Library()
      {
         fmt = NumberFormat.getInstance();
         fmt.setMinimumIntegerDigits(5);
         fmt.setGroupingUsed(false);         
         
         // create a single instance of an anonymous inner class and save it
         // in the instance member for future use
         alphaCompare = new Comparator()
         {
            // sorts in lexicographic order
            public int compare(Object o1, Object o2)
            {
               String s1 = (String)o1;
               return s1.compareToIgnoreCase((String)o2);
            }
         };         
         
         // create a single instance of an anonymous inner class and save it
         // in the instance member for future use
         matchCompare = new Comparator()
         {
            // sorts in descending order by the total number of matches,
            // within equal number of matches, sorts lexicographically
            // (this is needed to ensure that the equals behavior is
            // correct)
            public int compare(Object o1, Object o2)
            {
               Statistic s1 = (Statistic)stats.get(o1);
               Statistic s2 = (Statistic)stats.get(o2);
               
               int       m1 = s1.getAggregateCounter();
               int       m2 = s2.getAggregateCounter();
               
               if (m1 == m2)
               {
                  // sort equal matches lexicographically
                  return alphaCompare.compare(o1, o2);
               }
               else if (m1 > m2)
               {
                  return -2;
               }
               else
               {
                  return 2;
               }
            }
         };                  
      }
      
      /**
       * Manually adds a statistic to the list of watched stats, using a
       * short name as a key (so that it can be referenced again later)
       * and a long descriptive name that can be used in reports.
       * <p>
       * Matches can be counted for such statistics by using {@link #addMatch} 
       * with the short name as the key.  In addition, if the short name is
       * the same as the one that would be computed by 
       * {@link #addMatchMultiplexed}, then that method can be used as
       * well.
       * <p>
       * Reasons to use this manual approach include providing more control
       * over the report output (the description which is not available
       * in the multiplexed approach) and the ability to match on a
       * different or more complicated condition.  Another reason is that
       * once a statistic is loaded this way, even if there are no matches
       * an entry in the summary report and a detailed report will be
       * created for this statistic.  This fact can be useful if one wants  
       * to explicitly see the number of matches as 0 for a statistic that
       * isn't matched.
       *
       * @param    shortName
       *           A simple string that is used as a key to reference this
       *           statistic.
       * @param    longName
       *           The verbose description that will be used in reports.
       */
      public void addStatistic(String shortName, String longName)
      {
         Statistic s = (Statistic) stats.get(shortName);
         
         if (s == null)
            stats.put(shortName, new Statistic(longName));
      }
      
      /**
       * Automatically adds a statistic to the list of watched stats if it
       * doesn't already exist AND always adds a match to that statistic.
       * This function will derive a short name for the statistic as a key
       * (so that it can be referenced again later) and it will use the same
       * name as the long descriptive name that appears in reports.
       * <p>
       * The short name is derived using the AST provided description for the
       * token type of the matched AST. Uses the {@link #addStatistic} and
       * {@link #addMatch} functions as workers.
       *
       * @param    type
       *           The string describing the type of text dump to use in
       *           match processing.  See <code>addMatch</code>. 
       */
      public void addMatchMultiplexed(String type)
      throws IllegalArgumentException,
             AstException
      {
         Aast src = resolver.getSourceAst();
         
         // safety logic
         if (src == null)
         {
            throw new IllegalArgumentException("No source AST node!");
         }
                  
         String name = src.getDescriptiveTokenText();
         
         addStatistic(name, name);
         addMatch(name, type);
      }
      
      /**
       * Automatically adds a statistic to the list of watched stats if it
       * doesn't already exist AND always adds a match to that statistic.
       * <p>
       * Uses the {@link #addStatistic} and {@link #addMatch} functions as
       * workers.
       * <p>
       * The 'name' parameter eventually contributes to the filename of the
       * report. Thus, it can't contain slahes or otherwise the file creation
       * fails. For safety, this method translates the name and replaces all
       * slashes with underscores. This change only affects the filenames and
       * leaves the original names in the report.
       *
       * @param   name
       *          The name which is used (1) as a key to identify the
       *          statistic being matched, possibly translated, and 
       *          (2) as the long name of the statistic in the report summary.
       * @param   type
       *          The string describing the type of text dump to use in
       *          match processing. 
       */
      public void addMatchMultiplexed(String name, String type)
      throws AstException
      {
         String transName = name.replaceAll(File.separator, "_");
         addStatistic(transName, name);
         addMatch(transName, type);
      }
      
      /**
       * Adds a match to a specific statistic that already exists in our
       * list of watched stats. Each time a match is added, the line number,
       * column number and some text from the AST will be stored into a
       * file specific aggregator inside the statistic in question. Later,
       * this data is accessible for detailed reports and the counter of 
       * matches is used in detailed report headers and summary reports.
       * <p>
       * This function will use the short name for the statistic as a key
       * to find the proper statistic.
       * <p>
       * The text that will be dumped depends on the dump type specified by
       * the caller.  The following types are supported:
       * <p>
       * <ul>
       *    <li> 'simple' - just records the token text and the symbolic
       *         token name
       *    <li> 'simplePlus' - same as 'simple', but some summary statistics
       *         are added about descendents (total and immediate children)
       *         and the number of nodes between the AST and the root
       *         (depth)
       *    <li> 'lisp' - dumps the symbolic token names of the AST and all
       *         it's children in a lisp-like single line syntax (warning:
       *         this can get quite long)
       *    <li> 'parser' - duplicates the parser's human readable output
       *         format for the entire subtree rooted at the matched AST
       *         (warning: this can get long AND this can be a multi-line
       *         output)
       * </ul>
       * <p>
       * If this dump type is not one of the above specified type strings,
       * then the dump type string itself is used as the literal text of the
       * match.  This allows the caller to assert complete control over the
       * match text (and format of the contents).
       *
       * @param    shortName
       *           The key which is used to identify the statistic being
       *           matched.
       * @param    type
       *           The string describing the type of text dump to use in
       *           match processing. If this is not one of the above 
       *           specified strings, then this string is used as the
       *           literal text of the match.
       */
      public void addMatch(String shortName, String type)
      throws IllegalArgumentException,
             AstException
      {
         Statistic s = (Statistic) stats.get(shortName);
         
         // can't add a match to a non-existing stat
         if (s == null)
         {
            throw new IllegalArgumentException("Non-existent statistic " +
                                               shortName + "!");
         }
         
         Aast src = resolver.getSourceAst();
         
         // safety logic
         if (src == null)
         {
            throw new IllegalArgumentException("No source AST node!");
         }
         
         String filename = src.getFilename();
         
         if (filename == null)
         {
            throw new AstException("AstRegistry failure for ID " +
                                    src.getId().toString() + "!");
         }
         
         // obtain the text from the match, by default, dump current AST only
         StringBuilder sb = new StringBuilder(src.getText());
         sb.append(" [").append(src.getSymbolicTokenType()).append("]");
         String text = sb.toString();         
         
         // possible overrides by caller to dump differently
         if ("simple".equalsIgnoreCase(type))
         {
            // nothing to do except avoid the else block below
         }
         else if ("simplePlus".equalsIgnoreCase(type))
         {
            int numImmDesc = src.getNumImmediateChildren();
            int numDesc    = src.getNumChildren();
            int depth      = src.getDepth();
            
            sb.append(" <total ").append(Integer.toString(numDesc));
            sb.append("> <immed ").append(Integer.toString(numImmDesc));
            sb.append("> <depth ").append(Integer.toString(depth));
            sb.append(">");
            text = sb.toString();
         }
         else if ("lisp".equalsIgnoreCase(type))
         {
            // add the text of the immediate node
            sb = new StringBuilder( src.getText() );
            
            // dump full tree with lisp-like syntax
            text = sb.append(" ").append(src.toStringList()).toString();
         }
         else if ("parser".equalsIgnoreCase(type))
         {
            text = src.dumpTree();
            text = text.substring(0, text.length() - 1); 
         }
         else
         {
            // anything unrecognized is used as the literal text
            text = type;
         }         
 
         int lineno = src.getLine(); 
         int colno  = src.getColumn(); 
         
         // process artificial tokens differently
         if (lineno == 0)
         {
            Iterator iter = src.iterator();
            
            while (lineno == 0 && iter.hasNext())
            {
               // this is an "artificial token", but all such tokens are
               // usually associated with a line and column, based on the
               // first descendant which has a non-zero line number
               Aast child = (Aast) iter.next();
               
               lineno = child.getLine();
               colno  = child.getColumn();
            }
         }
         
         // map filename to ID if not already done
         Integer fid = file2ID.get(filename);
         if (fid == null)
         {
            fid = ++nextFileID;
            file2ID.put(filename, fid);
            id2File.put(fid, filename);
         }
         
         try
         {
            int sid = s.getID();
            stmtInsert.setInt(1, sid);
            stmtInsert.setInt(2, fid);
            stmtInsert.setInt(3, lineno);
            stmtInsert.setInt(4, colno);
            stmtInsert.setString(5, text);
            stmtInsert.execute();
            stmtInsert.clearParameters();
            
            connection.commit();
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException(
               "Error inserting statistic into database", exc);
         }
         
         s.match(filename,
                 conditionText == null ? resolver.getCurrentCondition()
                                       : conditionText,
                 lineno,
                 colno,
                 text);
      }
      
      /**
       * Sets default output directory in which all reports will be written.
       *
       * @param    dir
       *           The directory path in which to create all output files.
       *
       * @return   <code>true</code> if the directory member was changed, 
       *           <code>false</code> otherwise. 
       */
      public boolean setOutputDirectory(String dir)
      {
         if (dir == null)
            return false;
         
         StringBuilder sb = new StringBuilder(dir);
         
         if (!dir.endsWith(File.separator))
            sb.append(File.separator);
         
         this.dir = sb.toString();
         
         return true;
      }
      
      /**
       * Sets the source helper that provides the ability to create a source
       * file index.
       *
       * @param    srcHelper
       *           The helper instance.
       */
      public void setSourceFileHelper(SourceFileHelper srcHelper)
      {
         StatsWorker.this.srcHelper = srcHelper;
      }
      
      /**
       * Sets sort order used in summary reports.  The following sort orders
       * can be specified:
       * <p>
       * <ul>
       *    <li> 'matches' - descending order by number of total matches per
       *         statistic (this is the default sort order), within equal
       *         number of matches, sorts lexicographically (this is needed
       *         to ensure that the equals behavior is correct)
       *    <li> 'alpha' - normal alphanumeric (lexicographic) sort by
       *         statistic short name
       *    <li> 'insertion' - in the order the statistics were added; this
       *         is meaningless when using multiplexed statistics because the 
       *         order of insertion is essentially random, however in
       *         manually created reports, this order will maintain the
       *         exact order that the user manually inserted the statistics
       *         which gives complete control over ordering to the user
       * </ul>
       *
       * @param    sort
       *           The requested sort order.
       *
       * @return   <code>true</code> if the sort order was recognized, 
       *           <code>false</code> otherwise. 
       */
      public boolean setSortOrder(String sort)
      {
         if ("matches".equalsIgnoreCase(sort))
         {
            sortOrder = SORT_MATCHES;
         }
         else if ("alpha".equalsIgnoreCase(sort))
         {
            sortOrder = SORT_ALPHA;
         }
         else if ("insertion".equalsIgnoreCase(sort))
         {
            sortOrder = SORT_INSERTION;
         }
         else
         {
            return false;
         }
         
         return true;
      }
      
      /**
       * Sets the flag for generation of a source file cross-reference.
       * Setting this to <code>false</code> does not change the generation
       * of the source file targets themselves, it only inhibits the
       * generation of the top level HTML file with the cross reference
       * links.
       *
       * @param    xref
       *           If <code>true</code> a source file cross reference will be 
       *           generated for the list of files processed.
       */
      public void setGenerateFileXref(boolean xref)
      {
         file_xref = xref;
      }      
      
      /**
       * Sets the name of the source file cross-reference to be generated.
       *
       * @param    xref_name
       *           The name of the source file cross reference to be 
       *           generated for the list of files processed.
       */
      public void setXrefName(String xref_name)
      {
         this.xref_name = xref_name;
      }      
      
      /**
       * Sets the mutual exclusion flag for this report which (when
       * <code>true</code>) will cause the summary report to have extra
       * columns for percent of matches and percent of files.
       * <p>
       * This value is reset (cleared) by {@link #resetStatistics}.
       *
       * @param    mutex
       *           Value for the mutual exclusion flag.
       */
      public void setMutex(boolean mutex)
      {
         this.mutex = mutex;
      }      
      
      /**
       * Sets the global condition text override to the given text. If set, 
       * this is used in preference to the current condition as retrieved
       * from the pattern engine's symbol resolver. This will override the
       * condition text that appears in the footer of the details reports.
       * This is useful to set that text to something more relevant than the
       * latest expression being processed (since that latest expression
       * is usually not the expression that was used to find the match).
       * <p>
       * This value is reset (cleared) by {@link #resetStatistics}.
       *
       * @param    conditionText
       *           Override expression for use in the details report footer.
       */
      public void setConditionText(String conditionText)
      {
         this.conditionText = conditionText;
      }
      
      /**
       * Sets the number of matches per page in a detailed report.
       *
       * @param    pageSize
       *           Number of matches per page.
       */
      public void setPageSize(int pageSize)
      {
         this.pageSize = pageSize;
      }      
      
      /**
       * Clear (delete) all statistics and any report-specific data from
       * memory.
       */
      public void resetStatistics() 
      {
         conditionText = null;
         mutex         = false;
         current_cats  = null;
         stats         = new LinkedHashMap();
         filelist      = new LinkedHashSet();
      }
      
      /**
       * Enable/disable master report generation. While enabled, all summary
       * reports generated, will store data in a static list that allows the
       * creation of a master report index. When disabled, all such data is
       * discarded.
       *
       * @param    master
       *           <code>true</code> to enable master report generation,
       *           <code>false</code> to disable and clear all data.
       */
      public void setMasterReport(boolean master)
      {
         File mstr = new File(MASTER_INDEX_CACHE);
         
         if (master)
         {
            if (masterIndex == null)
            {
               if (mstr.exists() && mstr.isFile())
               {
                  try
                  {
                     restoreMappings(MASTER_INDEX_CACHE);
                     mstr.delete();
                  }
                  
                  catch (Exception exc)
                  {
                     exc.printStackTrace(System.err);
                  }
               }
               else
               {
                  masterIndex = new LinkedList();
               }
            }
         }
         else
         {
            // disable/clear
            masterIndex = null;
            categories  = null;
            
            if (mstr.exists() && mstr.isFile())
            {
               try
               {
                  mstr.delete();
               }
               
               catch (Exception exc)
               {
                  exc.printStackTrace(System.err);
               }
            }
         }
      }
      
      /**
       * Associate the given list of comma-separated categories with the
       * current report. This will cause the subsequent master report index
       * to be generated with a separate table for each category.
       *
       * @param    list
       *           The comma-separated list of categories to associate with
       *           the current report.
       */
      public void setMasterReportCategoryList(String list)
      {
         // split up the list
         String[] array = list.split(",", -1);
         
         for (int i = 0; i < array.length; i++)
         {
            if (array[i] != null && array[i].length() > 0)
            {
               setMasterReportCategory(array[i]);
            }
         }
      }
      
      /**
       * Associate the given text category with the current report. This will
       * cause the subsequent master report index to be generated with a
       * separate table for each category.
       *
       * @param    category
       *           The name of a category to associate with the current
       *           report.
       */
      public void setMasterReportCategory(String category)
      {
         if (categories == null)
         {
            categories = new LinkedHashMap();
         }
         
         if (current_cats == null)
         {
            current_cats = new LinkedHashSet();
         }
         
         categories.put(category, null);
         current_cats.add(category);
      }
      
      /**
       * Simple driver function to generate a summary report, a detailed
       * report for all statistics that have been gathered and a file index.
       *
       * @param    title
       *           The title used in the summary report.
       * @param    prefix
       *           Filename prefix for each output destination.  May be set
       *           to <code>null</code>.
       */
      public void generateAllReports(String title, String prefix)
      {
         StringBuilder sum = new StringBuilder();
         StringBuilder det = new StringBuilder();
         
         // fixup the prefix to specify a relative directory
         prefix = processDirPrefix(prefix);
         
         // generate a unique summary filename
         sum.append(prefix).append(SUM_NAME);
          
         // generate a unique prefix for each detailed report including the
         // relative directory path
         det.append(prefix).append(DET_PREFIX);
         
         generateSummaryReport(title, sum.toString(), DET_PREFIX);
         generateAllDetailedReportsWorker(det.toString());
         
         // generate the source file targets
         if (file_xref)
         {
            generateFileTargets(xref_name);
         }
         else
         {
            generateFileTargetsNoSummary();
         }
      }
      
      /**
       * Generates a detailed HTML report for every statistic that has been
       * matched.  The filename for the output file will be generated using
       * an optional prefix which will be prepended to a statistic-specific
       * short name and the string '.html' to generate a unique output
       * destination.  If set, the default output directory is honored.
       *
       * @param    prefix
       *           Filename prefix for each output destination.
       */
      public void generateAllDetailedReports(String prefix)
      {
         generateAllDetailedReportsWorker(processDirPrefix(prefix));
      }
      
      /**
       * For each filename, the core source files are written into various
       * html output and a cross-reference index is generated of file names
       * and the various types of viewable output.
       *
       * @param    output
       *           Output destination.
       */
      public void generateFileTargets(String output)
      {
         if (srcHelper == null)
         {
            throw new RuntimeException("Missing SourceFileHelper instance.");
         }
         
         IndentingWriter out = createStream(output, true);
         
         try
         {
            int num = filelist.size();
            
            StringBuilder sb = new StringBuilder("Source File Index");
            if (num > 1)
            {
               // if there are multiple files, tell how many
               sb.append(" (").append(num).append(" files)");
            }
            
            // write document header
            writeHeader(out, sb.toString());         
            
            Iterator fls = filelist.iterator();
   
            // only write the body if there are files in the list
            if (fls.hasNext())
            {
               // track if this row's background needs to be colorized or not
               boolean color = false;
               
               // write table header
               writeTableOpen(out);         
               writeRowStart(out, true, OLIVE);
               
               if (headings == null)
               {
                  headings = srcHelper.getHeadings();
               }
               
               for (int j = 0; j < headings.length; j++)
               {
                  writeField(out,
                             headings[j],
                             VALIGN_MIDDLE,
                             HALIGN_CENTER,
                             BOLD);            
               }
                          
               writeRowEnd(out);
               
               // print a line for each file
               while (fls.hasNext())
               {
                  String fname = (String) fls.next();
                  
                  // create the output subdir, copy/convert the files as
                  // needed
                  String[] targets = generateTargetOutput(fname);
                  
                  // bypass this row, if the files cannot be accessed
                  if (targets == null) 
                     continue;
                  
                  writeRowStart(out, color, null);
                                            
                  for (int x = 0; x < targets.length; x++)
                  {
                     // build descriptive name
                     int begin = 0;
                     int end   = targets[x].lastIndexOf('.');
                     
                     String tail = targets[x].substring(end);
                     
                     // if we copied to a txt or html file, we don't need the
                     // file extension in the link text 
                     if (!tail.equals(".txt") && !tail.equals(".html"))
                     {
                        end = targets[x].length();
                     }
                     
                     // we leave the path in for the first column
                     if (x > 0)
                     {
                        int idx = targets[x].lastIndexOf(File.separatorChar); 
                        begin = idx + 1;
                     }
                     
                     String descr = targets[x].substring(begin, end);
                     
                     writeField(out,
                                HtmlHelper.href(targets[x], descr, false),
                                VALIGN_MIDDLE,
                                HALIGN_LEFT,
                                NORMAL,
                                true);
                  }
                  
                  writeRowEnd(out);
                  
                  // toggle this value for the next row
                  color = !color;
               }         
               
               writeTableClose(out);
            }
            
            writeFooter(out, false, "");
         }
         
         finally
         {
            close(out);
         }
      }
      
      /**
       * For each filename, the core source files are written into various
       * html output and NO cross-reference index is generated.
       */
      public void generateFileTargetsNoSummary()
      {
         if (srcHelper == null)
         {
            throw new RuntimeException("Missing SourceFileHelper instance.");
         }
         
         int num = filelist.size();
         
         Iterator fls = filelist.iterator();

         // process each file
         while (fls.hasNext())
         {  
            // create the output subdir, copy/convert the files as needed
            generateTargetOutput((String) fls.next());               
         }         
      }
      
      /**
       * Simple driver function to generate the master index report, with
       * 1 entry for each summary report generated since the master report
       * support was reset (enabled for the first time or disabled and then
       * enabled).  To enable/disable, a separate call is needed to
       * {@link #setMasterReport}. This method emits the master index but 
       * does not affect any state.
       *
       * @param    fname
       *           Filename for the master index report.
       * @param    title
       *           The title text for the report or <code>null</code> for
       *           a generic default title.
       */
      public void generateMasterReport(String fname, String title)
      {
         if (masterIndex == null || masterIndex.size() == 0)
         {
            // nothing to do
            return;
         }
         
         IndentingWriter out = createStream(fname, true);
         
         try
         {
            // write document header
            writeHeader(out, title == null ? "Report Index" : title);         
            
            // only write the body if there are files in the list
            if (!masterIndex.isEmpty())
            {
               if (categories != null)
               {
                  // write the table of contents
                  writeTOC(out);
                  
                  Iterator cats = categories.keySet().iterator();
                  
                  // write a section of the master report for each category
                  while (cats.hasNext())
                  {
                     String next = (String) cats.next();
                     
                     // find the anchor name for the category
                     String anchor = (String) categories.get(next);
                     
                     // create the named anchor and insert it in the body
                     write(out, HtmlHelper.namedAnchor(anchor, ""));
                     
                     // write a table name for the category
                     writeHeading(out, 3, next);
                     
                     // write a table with only those reports associated with
                     // the next category
                     writeMasterReportTable(out, next);
                     
                     // separate each category
                     writeSeparator(out);
                  }
                  
                  // create the named anchor and insert it in the body
                  String ara = HtmlHelper.namedAnchor(allReportsAnchor, "");
                  write(out, ara);
                  
                  // write a table name for the "all reports" case
                  writeHeading(out, 3, ALL_RPTS);
               }
               
               // always write the "all reports" table
               writeMasterReportTable(out, null);
               
               if (categories != null)
               {
                  // separate each category
                  writeSeparator(out);
               }
            }
            
            writeFooter(out, true, "");
         }
         
         finally
         {
            close(out);
         }
      }
      
      /**
       * Generates a summary report for all statistics matched in this rule
       * set. Each summary report contains a header with a title (passed in
       * by the caller) the total number of matches for all statistics and
       * the total number of files in which any match was found. The body of
       * the report contains a summary line for each statistic matched, with
       * the number of matches for that statistic and the number of files in
       * which that statistic was matched.
       *
       * @param    title
       *           Title for header of the report.
       * @param    output
       *           Output destination.
       * @param    prefix
       *           Optional prefix for creating links to the detailed
       *           report names.
       */
      public void generateSummaryReport(String title,
                                        String output,
                                        String prefix)
      {
         int totalMatches = 0;
         int totalFiles   = 0;
         
         IndentingWriter out = createStream(output, true);

         try
         {
            Iterator iter = stats.keySet().iterator();
            
            // aggregate totals for header
            while (iter.hasNext())
            {
               Statistic s = (Statistic) stats.get((String)iter.next());
               
               totalMatches += s.getAggregateCounter();
            }
            
            Iterator fls = filelist.iterator();
            
            // we can't use Statistic.matchFiles because the counter will
            // potentially count the same file multiple times when you have
            // multiple statistics         
            while (fls.hasNext())
            {
               String fname = (String) fls.next();
               
               iter = stats.keySet().iterator();
               
               // check if any of the statistics had a match in this file, if
               // so then our total files count will increase by 1
               while (iter.hasNext())
               {
                  Statistic s = (Statistic) stats.get((String)iter.next());
                  
                  // if any of the statistics have a match in this file, we
                  // add this file to the total and then break out
                  if (s.getCounterFor(fname) > 0)
                  {
                     totalFiles++;
                     break;
                  }               
               }
            }
            
            // write document header
            writeStandardHeader(out, title, totalMatches, totalFiles);
            
            // store master report index data if needed
            if (masterIndex != null)
            {
               MasterReportEntry mre = new MasterReportEntry();
               mre.descr      = title;
               mre.url        = output;
               mre.matches    = totalMatches;
               mre.files      = totalFiles;
               mre.categories = current_cats;
               
               masterIndex.add(mre);
            }
            
            // track if this row's background needs to be colorized or not
            boolean color = false;
            
            // write table header
            writeTableOpen(out);         
            writeRowStart(out, true, OLIVE);
            
            writeField(out,
                       "Condition",
                       VALIGN_MIDDLE, 
                       HALIGN_CENTER,
                       BOLD);
            writeField(out,
                       "# Matches",
                       VALIGN_MIDDLE,
                       HALIGN_CENTER,
                       BOLD);
            if (mutex)
            {
               writeField(out,
                          "% Matches",
                          VALIGN_MIDDLE,
                          HALIGN_CENTER,
                          BOLD);
            }
            writeField(out,
                       "# Files",
                       VALIGN_MIDDLE,
                       HALIGN_CENTER,
                       BOLD);            
            if (mutex)
            {
               writeField(out,
                          "% Files",
                          VALIGN_MIDDLE,
                          HALIGN_CENTER,
                          BOLD);
            }
            
            writeRowEnd(out);
                     
            TreeSet ts;
                     
            // obtain an iterator to the keys, with the requested sort order
            // (defaults to SORT_MATCHES)
            switch (sortOrder)
            {
               case SORT_INSERTION:
                  // returns in the order in which the statistics were added
                  iter = stats.keySet().iterator();
                  break;
               case SORT_ALPHA:
                  // alpha sort by statistic short name
                  ts = new TreeSet(alphaCompare);
                  ts.addAll(stats.keySet());
                  iter = ts.iterator();
                  break;
               case SORT_MATCHES:               
               default:         
                  // descending sort by statistic aggregate count 
                  ts = new TreeSet(matchCompare);
                  ts.addAll(stats.keySet());
                  iter = ts.iterator();
                  break;
            }  
            
            // print summary entries, one line per statistic
            while (iter.hasNext())
            {
               String    shortName = (String) iter.next();
               Statistic s         = (Statistic) stats.get(shortName);
               
               int mcounter = s.getAggregateCounter();
               int fcounter = s.getMatchFiles();
               
               String mstring = Integer.toString(mcounter);
               String fstring = Integer.toString(fcounter);
               
               // build link to the details report
               String url = generateSafeFilename(prefix, shortName, ".html");
               
               writeRowStart(out, color, null);
               
               // emit the fields for the new row
               writeField(out,
                          HtmlHelper.href(url, s.getDescr(), false),
                          VALIGN_MIDDLE,
                          HALIGN_LEFT,
                          NORMAL, 
                          true);
                          
               writeField(out,
                          mstring,
                          VALIGN_MIDDLE,
                          HALIGN_RIGHT,
                          NORMAL);
                          
               if (mutex)
               {
                  writeField(out,
                             formatAsPercent(mcounter, totalMatches),
                             VALIGN_MIDDLE,
                             HALIGN_RIGHT,
                             NORMAL);
               }
               
               writeField(out,
                          fstring,
                          VALIGN_MIDDLE,
                          HALIGN_RIGHT,
                          NORMAL);
                          
               if (mutex)
               {
                  writeField(out,
                             formatAsPercent(fcounter, totalFiles),
                             VALIGN_MIDDLE,
                             HALIGN_RIGHT,
                             NORMAL);
               }
               
               writeRowEnd(out);
               
               // toggle this value for the next row
               color = !color;
            }
   
            writeTableClose(out);
            writeFooter(out, true, calcRelativePath(output));
         }
         
         finally
         {
            close(out);
         }
      }
      
      /**
       * Generates a detailed report for a single named statistic. Each
       * detailed report contains a header with a title (based on the
       * statistic's description), the number of matches and the number of
       * files in which a match was found. The body of the report contains a
       * line for each instance of a match, with a simple index referencing
       * the filename, a line number, a column number and the text of the
       * match.
       *
       * @param    shortName
       *           Identifies the statistic to print.
       * @param    output
       *           Output destination.
       */
      public void generateDetailedReport(String shortName, String output)
      {
         IndentingWriter out     = createStream(output, true);
         String          relPath = calcRelativePath(output);
         
         try
         {
            Statistic stat = (Statistic) stats.get(shortName);
            
            // can't add a match to a non-existent stat
            if (stat == null)
            {   
               throw new IllegalArgumentException("No such statistic named " +
                                                  shortName + "!");
            }
            
            writeReportHeader(out, stat);
            
            int total = stat.getAggregateCounter();
            
            if (total > pageSize)
            {
               // calculate base filename (including subdirectories)
               String baseName = output.substring(0, output.lastIndexOf('.'));
               
               // calculate the base link name (prev/toc/next links for each
               // page)
               int idx = baseName.lastIndexOf(File.separatorChar);
               String baseLink = (idx == -1) ? baseName
                                             : baseName.substring(++idx);
               
               // calculate the max number of digits in the number of matches
               int max = String.format("%d", total).length();
               
               // create filename format spec with proper width
               String spec =
                  String.format("%%s_%%0%1$dd_through_%%0%1$dd.html", max);
               
               // create the TOC description spec
               String dspec =
                  String.format("%%0%1$dd through %%0%1$dd", max);
                                
               // create the title spec
               String tspec =
                  String.format("%%s (matches %%0%1$dd through %%0%1$dd)",
                                max);
                  
               // calculate the table of contents link
               idx = output.lastIndexOf(File.separatorChar);
               String toc = (idx == -1) ? output : output.substring(++idx);
               
               String prev = null;
               String curr = null;
               String next = null;
               
               writeDetailsPageTOCHeader(out);
               
               // process file by file
               Iterator filelist = stat.getFileIterator();
               
               // keep track of partial match sets in process between pages
               PageState ps = null;
               
               int page = 0;
                                             
               // break our output into multiple pages
               for (int start = 1; start <= total; start += pageSize)
               {
                  // calculate page number
                  page++;
                  
                  // calculate the end match number for this page and the
                  // start/end match numbers for the next page (lookahead)
                  int end       = start + pageSize - 1;
                  int nextStart = end + 1;
                  int nextEnd   = nextStart + pageSize - 1;
                  
                  if (end > total)
                  {
                     end = total;
                     
                     nextStart = -1;
                     nextEnd   = -1;
                  }
                  else if (nextEnd > total) 
                  {
                     nextEnd = total;
                  }
                  
                  // calculate the next page filename
                  String fname = String.format(spec, baseName, start, end);
                  
                  // calculate the links
                  prev = curr;
                  curr = next;
                  next = String.format(spec, baseLink, nextStart, nextEnd);
                  
                  // is this the first time through?
                  if (curr == null)
                  {
                     // pre-load the curr link
                     curr = String.format(spec, baseLink, start, end);
                  }
                  
                  // is this the last time through?
                  if (nextStart == -1)
                  {
                     // disable the next link
                     next = null;
                  }
                  
                  // calculate the title
                  String title = String.format(tspec,
                                               stat.getDescr(),
                                               start,
                                               end);
                  
                  // write the current page of output
                  ps = writeReportDetailsPage(fname,
                                              relPath,
                                              stat,
                                              filelist,
                                              ps,
                                              title,
                                              end - start + 1,
                                              prev,
                                              toc,
                                              next);
                                              
                  // calculate the TOC description
                  String descr = String.format(dspec, start, end);
                  
                  // write the table of contents entry
                  writeDetailsPageTOCEntry(out,
                                           page,
                                           curr,
                                           descr,
                                           ps.start,
                                           ps.end);
               }
               
               writeDetailsPageTOCEnd(out);
            }
            else
            {
               // simple 1 page report
               writeReportDetails(out, stat, relPath);
            }
            
            writeFooter(out, true, relPath);
         }
         
         finally
         {
            close(out);
         }
      }
      
      /**
       * Generates a custom report using the same formatting as for standard
       * reports, but using data provided by the caller. Each custom report
       * contains a header with a title. The body of the report contains a
       * table with a header specified by the header list of strings (1
       * column per header string).  This table will also have a row for each
       * values list contained in the given list of rows. 
       *
       * @param    output
       *           The simple filename of the report (should not include the
       *           output directory since that will be prepended as needed).
       * @param    title
       *           The report's title.
       * @param    headers
       *           An ordered list of each column's label. This must be a
       *           list with elements of type <code>String</code>.  The
       *           number of elements will be the number of columns created
       *           for the table.
       * @param    rows
       *           An ordered list containing all row data.  This is a list
       *           which contains lists of each column's data for each row in
       *           the table.  This must be a list with elements that are
       *           lists with elements of type <code>String</code>.  The
       *           number of elements in each contained list must be the same
       *           as the number of elements of the <code>headers</code> list.
       * @param    totalMatches
       *           The total number of matches to be shown in the master
       *           report index for this report.
       * @param    totalFiles
       *           The total number of files (that had any matches) to be
       *           shown in the master report index for this report.
       */
      public void generateCustomReport(String output,
                                       String title,
                                       List   headers,
                                       List   rows,
                                       int    totalMatches,
                                       int    totalFiles)
      {
         IndentingWriter out = createStream(output, true);
         
         try
         {
            writeHeader(out, title);
            writeCustomReportDetails(out, headers, rows);
            writeFooter(out, false, "");
            
            // store master report index data if needed
            if (masterIndex != null)
            {
               MasterReportEntry mre = new MasterReportEntry();
               mre.descr      = title;
               mre.url        = output;
               mre.matches    = totalMatches;
               mre.files      = totalFiles;
               mre.categories = current_cats;
               
               masterIndex.add(mre);
            }
         }
         
         finally
         {
            close(out);
         }
      }
      
      /**
       * Given an XML filename and a distribution type, load statistics from
       * an XML document and produce a map of all distributions of that type.
       * The map's keys are the distribution names; its values are the
       * distributions themselves.
       * <p>
       * Each distribution is represented as a <code>SortedMap</code>, where
       * each map key represents a "bucket" value, and each map value
       * represents the the number of occurrences of that value in the
       * distribution.
       *
       * @param   filename
       *          Name of the file containing the XML document to be read.
       * @param   type
       *          Type of distribution to collect; this corresponds with the
       *          "type" attribute of the "distribution" element within the
       *          XML document.
       *
       * @return  An unmodifiable map of unmodifiable distribution maps.
       *
       * @throws  RuntimeException
       *          if any error occurs loading the distribution data from the
       *          specified XML file.
       *
       * @see     StatisticsHelper#loadDistributions
       */
      public Map loadDistributions(String filename, String type)
      {
         return StatisticsHelper.loadDistributions(filename, type);
      }
      
      /**
       * Calculate the total population size of a distribution.
       *
       * @param   distrib
       *          Map of data values to numbers of occurrences.
       *
       * @return  Cumulative population of all distribution buckets.
       *
       * @see     StatisticsHelper#population
       */
      public long population(Map distrib)
      {
         return StatisticsHelper.population(distrib);
      }
      
      /**
       * Determine the maximum value in a distribution.
       *
       * @param   distrib
       *          Map of data values to number of occurrences. Keys must
       *          implement <code>Comparable</code>.
       *
       * @return  Maximum value.
       *
       * @see     StatisticsHelper#maximum
       */
      public Object maximum(Map distrib)
      {
         return StatisticsHelper.maximum(distrib);
      }
      
      /**
       * Determine the minimum value in a distribution.
       *
       * @param   distrib
       *          Map of data values to number of occurrences. Keys must
       *          implement <code>Comparable</code>.
       *
       * @return  Minimum value.
       *
       * @see     StatisticsHelper#minimum
       */
      public Object minimum(Map distrib)
      {
         return StatisticsHelper.minimum(distrib);
      }
      
      /**
       * Determine the median value in a distribution.
       *
       * @param   distrib
       *          Map of data values to number of occurrences. Keys must
       *          implement <code>Comparable</code>.
       *
       * @return  Median value.
       *
       * @see     StatisticsHelper#median
       */
      public double median(Map distrib)
      {
         return StatisticsHelper.median(distrib);
      }
      
      /**
       * Determine the weighted average value in a distribution.
       *
       * @param   distrib
       *          Map of data values to number of occurrences. Keys must
       *          be instances of <code>Number</code>.
       *
       * @return  Mean value.
       *
       * @see     StatisticsHelper#mean
       *
       * @throws  ClassCastException
       *          if a non-numeric key is encountered in <code>distrib</code>.
       */
      public double mean(Map distrib)
      {
         return StatisticsHelper.mean(distrib);
      }
      
      /**
       * Determine the standard deviation of a distribution.
       *
       * @param   distrib
       *          Map of data values to number of occurrences. Keys must
       *          be instances of <code>Number</code>.
       *
       * @return  Standard deviation.
       *
       * @see     StatisticsHelper#standardDeviation
       *
       * @throws  ClassCastException
       *          if a non-numeric key is encountered in <code>distrib</code>.
       */
      public double standardDeviation(Map distrib)
      {
         return StatisticsHelper.standardDeviation(distrib);
      }
      
      /**
       * Creates the directory tree to the target if needed and then copies
       * the source to the target if needed, during the copy, the input file
       * is written to an HTML output file with named anchors created for
       * each line.  The result is a file that can be linked to on a line by
       * line basis.
       *
       * @param    src
       *           Source filename.
       * @param    tar
       *           Target filename.
       *
       * @return   <code>true</code> if the file was copied successfully OR
       *           if the file didn't need to be copied, <code>false</code>
       *           on any failure.
       */
      public boolean copyToHtml(String src, String tar)
      {  
         // try to create the directory
         if (!Utils.createPathForTarget(tar))
         {
            // dir creation failed
            return false;
         }
         
         // bypass this if the target file exists and has a later timestamp
         // than the source file
         if (Utils.isLaterVersion(src, tar))
         {
            // nothing to do
            return true;
         }
         
         // the output directory is already prepended, so create a stream
         // using the exact target name
         IndentingWriter out = createStream(tar, false);         
         
         try
         {
            // calculate our document title
            int begin = tar.lastIndexOf(File.separatorChar) + 1;
            int end   = tar.lastIndexOf('.');
            
            String title = tar.substring(begin, end);         
            
            // initialize the header and document title
            writeSimpleHeader(out, title);
            
            // open our <pre> tag
            int level = out.getIndent();
            out.setIndent(0);
            write(out, HtmlHelper.startPre());
            
            try
            {
               BufferedReader in = new BufferedReader(new FileReader(src));
               
               int    line = 1;
               String next = in.readLine();            
               
               while (next != null)
               {
                  // entity conversion
                  String safe = HtmlHelper.filter(next);
                  
                  // format the line number
                  String        lineno = fmt.format(line);
                  StringBuilder sb     = new StringBuilder(lineno);
                  sb.append("  ");
                  sb.append(safe);
                  
                  // write the anchor
                  String na = HtmlHelper.namedAnchor(lineno,
                                                     sb.toString(),
                                                     false);
                  write(out, na);
                  
                  // prepare for next line
                  line++;
                  next = in.readLine();
               }
            }
            catch (IOException io)
            {
               return false;
            }
            
            write(out, HtmlHelper.endPre());
            out.setIndent(level);
            
            writeSimpleFooter(out);
         }
         
         finally
         {
            close(out);
         }
         
         return true;
      }
      
      /**
       * Save the name mapping data into the given file using serialization. 
       *
       * @param    filename
       *           The file in which to write the name mappings.
       */
      public void saveMappings(String filename)
      throws FileNotFoundException,
             IOException
      {
         FileOutputStream   fos = new FileOutputStream(filename, false);
         ObjectOutputStream oos = new ObjectOutputStream(fos);
         
         oos.writeObject(masterIndex);
         oos.writeObject(categories);
         oos.close();
      }
      
      /**
       * Restore the name mapping data from the given file using
       * serialization. 
       *
       * @param    filename
       *           The file from which to read the name mappings.
       */
      public void restoreMappings(String filename)
      throws FileNotFoundException,
             IOException,
             ClassNotFoundException
      {
         FileInputStream   fis = new FileInputStream(filename);
         ObjectInputStream ois = new ObjectInputStream(fis);
         
         masterIndex = (LinkedList) ois.readObject();
         categories  = (LinkedHashMap) ois.readObject();
         ois.close();
      }
      
      /**
       * Generates a detailed HTML report for every statistic that has been
       * matched.  The filename for the output file will be generated using
       * an optional prefix which will be prepended to a statistic-specific
       * short name and the string '.html' to generate a unique output
       * destination.  If set, the default output directory is honored.
       *
       * @param    prefix
       *           Filename prefix for each output destination. This may
       *           include a relative path segments so long as they don't 
       *           duplicate those in the default output directory.
       */
      private void generateAllDetailedReportsWorker(String prefix)
      {
         Iterator iter = stats.keySet().iterator();
         
         while (iter.hasNext())
         {
            String shortName = (String) iter.next();
            String url = generateSafeFilename(prefix, shortName, ".html");
            
            generateDetailedReport(shortName, url);
         }
      }
      
      /**
       * Creates a filename that can be safely used in both the file system
       * and as a URL.
       *
       * @param    prefix
       *           The text for the beginning of the name.
       * @param    name
       *           The base name itself.
       * @param    suffix
       *           The text for the end of the name.
       *
       * @return   The encoded filename/URL.
       */
      private String generateSafeFilename(String prefix,
                                          String name,
                                          String suffix)
      {
         StringBuilder sb = new StringBuilder(prefix);
         sb.append(name);
         
         String input = sb.toString().toLowerCase();
         
         StringBuilder fname = new StringBuilder();
         
         char last = 0;
         int  i;
            
         for (i = 0; i < input.length() && i < FILENAME_LIMIT; i++)
         {
            char ch = input.charAt(i);
            
            if (Character.isWhitespace(ch))
            {
               if (last != '_')
               {
                  ch = '_';
               }
               else
               {
                  // compress contiguous whitespace
                  continue;
               }
            }
            else
            {
               boolean norm = Character.isLetterOrDigit(ch);
               boolean sep  = (ch == File.separatorChar);
               
               // is it a valid character?
               if (!norm && !sep && ch != '.' && ch != '-' && ch != '_')
               {
                  // drop the character (hopefully it is not needed to make
                  // the name unique)
                  continue;
               }
            }
            
            last = ch;
            fname.append(ch);
         }
         
         if (i == FILENAME_LIMIT - 1)
         {
            // we may have truncated, now add a unique index to make sure
            // this is differentiated
            fname.append(String.format("_%08x", counter++));
         }
         
         fname.append(suffix);
         
         return fname.toString();
      }
      
      /**      
       * Write a table of master report results to the given document. Each
       * report entry gets a separate row in the table.
       *
       * @param    out
       *           The output destination.
       * @param    category
       *           Only write report entries that are associated with this
       *           category.  If <code>null</code>, write all possible report
       *           entries.
       */
      private void writeMasterReportTable(IndentingWriter out,
                                          String          category)
      {
         Iterator rpts = masterIndex.iterator();
         
         // track if this row's background needs to be colorized or not
         boolean color = false;
         
         // write table header
         writeTableOpen(out);         
         writeRowStart(out, true, OLIVE);
         
         writeField(out,
                    "Report Name",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);            
         writeField(out,
                    "# Matches",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "# Files",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
                    
         writeRowEnd(out);
         
         // print a line for each file
         while (rpts.hasNext())
         {
            MasterReportEntry mre = (MasterReportEntry) rpts.next();
            
            // honor given category if necessary
            if (category != null        &&
                mre.categories != null  &&
                !mre.categories.contains(category))
            {
               // skip this report entry
               continue;
            }
            
            writeRowStart(out, color, null);
                                      
            String ms = Integer.toString(mre.matches);
            String fs = Integer.toString(mre.files);
                       
            writeField(out,
                       HtmlHelper.href(mre.url, mre.descr, false),
                       VALIGN_MIDDLE,
                       HALIGN_LEFT,
                       NORMAL,
                       true);
            writeField(out,
                       ms,
                       VALIGN_MIDDLE,
                       HALIGN_LEFT,
                       NORMAL);            
            writeField(out,
                       fs,
                       VALIGN_MIDDLE,
                       HALIGN_LEFT,
                       NORMAL);            
            
            writeRowEnd(out);
            
            // toggle this value for the next row
            color = !color;
         }         
         
         writeTableClose(out);
      }
      
      /**    
       * Creates a subdirectory in the output directory to match the
       * source path, then copies the original source file (unchanged),
       * the preprocessed file (written in HTML with anchors for line number
       * linking), the lexer output (unchanged) and the parser output
       * (unchanged).
       *
       * @param    filename
       *           The original source file.
       *
       * @return   The list of the exact target filenames or <code>null<code>
       *           if the files cannot be accessed.
       */
      private String[] generateTargetOutput(String filename)
      {
         CopySpec[] specs = srcHelper.getSpecifications(filename, dir);
         
         // process each file to ensure it is available as a target
         for (int i = 0; i < specs.length; i++)
         {
            String        src  = specs[i].getSource();
            String        tar  = specs[i].getTarget();
            CopySpec.Type type = specs[i].getOperation();
            
            if (type == CopySpec.Type.BINARY)
            {
               Utils.simpleCopy(src, tar);
            }
            else if (type == CopySpec.Type.TXT_TO_HTML)
            {
               copyToHtml(src, tar);
            }
         }
         
         String[] targets = new String[specs.length];
         
         // remove the prepended output dir before returning the targets
         int len = dir.length();
         
         for (int x = 0; x < specs.length; x++)
         {
            targets[x] = specs[x].getTarget().substring(len);
         }
         
         return targets;
      }
      
      /**
       * Formats the <code>value</code> as a percent of <code>max</code>.
       *
       * @param    value
       *           The value which is considered a fraction of the possible
       *           maximum.
       * @param    max
       *           The maximum possible value.
       *
       * @return   The resulting percentage. 
       */
      private String formatAsPercent(int value, int max)
      {
         double percent = (double) value / (double) max;   
         return NumberFormat.getPercentInstance().format(percent);
      }
      
      /**
       * Prints a header of a detailed report on a specific statistic.
       *
       * @param    out
       *           The output destination.
       * @param    stat
       *           The statistic to print.
       */
      private void writeReportHeader(IndentingWriter out, Statistic stat)
      throws IllegalArgumentException
      {
         writeStandardHeader(out,
                             stat.getDescr(),
                             stat.getAggregateCounter(),
                             stat.getMatchFiles());
      }
         
      /**
       * Prints a header of a summary or detailed report.
       *
       * @param    out
       *           The output destination.
       * @param    descr
       *           The title text.
       * @param    matches
       *           The total number of matches for this statistic or for the
       *           set of statistics.
       * @param    files
       *           The total number of files in which matches were found.
       */
      private void writeStandardHeader(IndentingWriter out,
                                       String          descr,   
                                       int             matches,
                                       int             files)
      {  
         StringBuilder sb = new StringBuilder(descr);
         sb.append(" (").append(matches).append(" in ");
         sb.append(files).append(" files)");
         
         writeHeader(out, sb.toString());
      }
      
      /**
       * Write the header row of the table of contents for a multi-page
       * details report.
       *
       * @param    out
       *           The output destination.
       */
      private void writeDetailsPageTOCHeader(IndentingWriter out)
      {
         writeTableOpen(out);
         
         writeRowStart(out, true, OLIVE);
         
         writeField(out,
                    "Page #", 
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "Link",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "Start Filename",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "End Filename",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
                    
         writeRowEnd(out);
      }
      
      /**
       * Write an entry in the table of contents for a multi-page details
       * report.
       *
       * @param    out
       *           The output destination.
       * @param    page
       *           Page number for the entry.
       * @param    link
       *           The link to create as an HREF.
       * @param    descr
       *           Description for the HREF.
       * @param    start
       *           Filename associated with the first entry in that page.
       * @param    end
       *           Filename associated with the last entry in that page.
       */
      private void writeDetailsPageTOCEntry(IndentingWriter out,
                                            int             page,
                                            String          link,
                                            String          descr,
                                            String          start,
                                            String          end)
      {
         writeRowStart(out, ((page % 2) != 0), null);
         
         writeField(out,
                    String.format("%d", page),
                    VALIGN_MIDDLE,
                    HALIGN_LEFT,
                    NORMAL);
         writeField(out,
                    HtmlHelper.href(link, descr, false), 
                    VALIGN_MIDDLE,
                    HALIGN_LEFT,
                    NORMAL,
                    true);
         writeField(out,
                    start,
                    VALIGN_MIDDLE,
                    HALIGN_LEFT,
                    NORMAL);
         writeField(out,
                    end,
                    VALIGN_MIDDLE,
                    HALIGN_LEFT,
                    NORMAL);
         
         writeRowEnd(out);
      }
      
      /**
       * End the table of contents for a multi-page details report.
       *
       * @param    out
       *           The output destination.
       */
      private void writeDetailsPageTOCEnd(IndentingWriter out)
      {
         writeTableClose(out);
      }
            
      /**
       * Writes the body of a detailed report for a single statistic.
       *
       * @param    out
       *           The output destination.
       * @param    stat
       *           The statistic to be reported upon.
       * @param    relPath
       *           The relative path to the "root" directory in which the
       *           source filenames exist. 
       */
      private void writeReportDetails(IndentingWriter out,
                                      Statistic       stat,
                                      String          relPath)
      throws IllegalArgumentException
      {
         indexAndAnalyze();
         
         try
         {
            int lastFID = -1;
            int matchCount = 0;
            String file = null;
            boolean wroteHeader = false;
            
            int sid = stat.getID();
            
            stmtQueryByStat.setInt(1, sid);
            ResultSet rs = stmtQueryByStat.executeQuery();
            
            // track if this row's background needs to be colorized or not
            boolean color = false;
            
            while (rs.next())
            {
               if (!wroteHeader)
               {
                  color = false;
                  writeDetailsTableHeader(out);
                  wroteHeader = true;
               }
               
               int fid = rs.getInt("file");
               
               if (fid != lastFID)
               {
                  lastFID = fid;
                  file = id2File.get(fid);
                  matchCount = stat.getCounterFor(file);
               }
               
               // bypass empty files
               if (matchCount == 0)
               {
                  continue;
               }
               
               int lineno = rs.getInt("line");
               int colno = rs.getInt("col");
               String text = rs.getString("text");
               Match match = new Match(lineno, colno, text);
               
               writeDetailsTableEntry(out, match, relPath, file, color);
               
               // toggle this value for the next row
               color = !color;
            }
            
            if (wroteHeader)
            {
               writeTableClose(out);
               writeCondition(out, stat);
               writeSeparator(out);
            }
            
            connection.commit();
         }
         catch (SQLException exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error writing report details", exc);
         }
      }
      
      /**
       * Write a single page of a multi-page details report.
       *
       * @param    output
       *           The page-specific output filename.
       * @param    relPath
       *           The relative path to the "root" directory in which the
       *           source filenames exist. 
       * @param    stat
       *           The statistic to be reported upon.
       * @param    filelist
       *           File list being processed.
       * @param    ps
       *           Partial state of output for a given file which is being
       *           output to more than one page. <code>null</code> if there
       *           is no state in process from the last file or if there was
       *           no file previously processed.
       * @param    title
       *           Title text for the page.
       * @param    num
       *           The number of entries (matches) to write in the table.
       * @param    prevLink
       *           Link to the previous page. <code>null</code> for the
       *           first page of output.
       * @param    tocLink
       *           Link to the table of contents.
       * @param    nextLink
       *           Link to the next page. <code>null</code> for the
       *           last page of output.
       *
       * @return   Any partial state that needs to be kept between pages
       *           is returned.
       */
      private PageState writeReportDetailsPage(String    output,
                                               String    relPath,
                                               Statistic stat,
                                               Iterator  filelist,
                                               PageState ps,
                                               String    title,
                                               int       num,
                                               String    prevLink,
                                               String    tocLink,
                                               String    nextLink)
      {
         IndentingWriter out = createStream(output, true);
         
         Iterator mset = (ps == null) ? null : ps.mset;
         String   file = (ps == null) ? null : ps.file;
         
         String startFile = null;
         String endFile   = null;
         
         try
         {
            // track if a row's background needs to be colorized or not
            boolean color = false;
            
            writeHeader(out, title);
            writePageLinks(out, prevLink, tocLink, nextLink);
            writeDetailsTableHeader(out);
            
            while (num > 0)
            {
               if (mset == null || !mset.hasNext())
               {
                  while (filelist.hasNext())
                  {
                     file = (String) filelist.next();
                     
                     // bypass empty files
                     if (stat.getCounterFor(file) > 0)
                        break;
                  }
                  
                  Long fidx = mgr.getTreeId(file);
                  mset = getMatchesFor(stat, file);            
               }
               
               while (mset.hasNext() && num > 0)
               {
                  Match next = (Match) mset.next();
                  num--;
                  
                  if (startFile == null)
                  {
                     startFile = file;
                  }
                  
                  endFile = file;
                  
                  writeDetailsTableEntry(out, next, relPath, file, color);
                  
                  // toggle this value for the next row
                  color = !color;
               }
            }
            
            writeTableClose(out);
            writePageLinks(out, prevLink, tocLink, nextLink);
            writeCondition(out, stat);
            writeSeparator(out);
            writeFooter(out, true, relPath);
         }
         
         finally
         {
            close(out);
         }
         
         if (ps == null)
         {
            ps = new PageState();
         }
         
         ps.mset  = (mset != null && !mset.hasNext()) ? null : mset;
         ps.file  = file;
         ps.start = startFile;
         ps.end   = endFile;
         
         return ps;
      }
      
      /**
       * Get an iterator for all statistics of the given type and filename.
       *
       * @param    stat
       *           The statistic to be reported upon.
       * @param    filename
       *           The file to be reported upon.
       * 
       * @return   See above.
       * @throws   RuntimeException if there were error computing matches.
       */
      private Iterator<Match> getMatchesFor(Statistic stat, String filename)
      {
         final int sid = stat.getID();
         final int fid = file2ID.get(filename);
         final ResultSet rs;
         
         indexAndAnalyze();
         
         try
         {
            stmtQueryByFile.setInt(1, sid);
            stmtQueryByFile.setInt(2, fid);
            rs = stmtQueryByFile.executeQuery();
         }
         catch (Exception exc)
         {
            try
            {
               connection.rollback();
            }
            catch (SQLException sqle)
            {
               // ignore
            }
            
            throw new RuntimeException("Error retrieving matches", exc);
         }
         
         return (new Iterator<Match>()
         {
            private Match next = null;
            
            public boolean hasNext()
            {
               maybeFetchNext();
               
               return (next != null);
            }
            
            public Match next()
            {
               maybeFetchNext();
               
               if (next == null)
               {
                  throw new NoSuchElementException();
               }
               
               try
               {
                  return next;
               }
               finally
               {
                  // reset for next pass
                  next = null;
               }
            }
            
            public void remove()
            {
               throw new UnsupportedOperationException();
            }
            
            private void maybeFetchNext()
            {
               if (next != null)
               {
                  return;
               }
               
               try
               {
                  if (rs.next())
                  {
                     int lineno = rs.getInt("line");
                     int colno = rs.getInt("col");
                     String text = rs.getString("text");
                     next = new Match(lineno, colno, text);
                  }
                  else
                  {
                     next = null;
                     
                     // we've reached the end;  now commit
                     connection.commit();
                  }
               }
               catch (SQLException exc)
               {
                  try
                  {
                     connection.rollback();
                  }
                  catch (SQLException sqle)
                  {
                     // ignore
                  }
                  
                  throw new RuntimeException("Error fetching match record",
                                             exc);
               }
            }
         });
      }
      
      /**
       * Write the links for the current page of a multi-page details report.
       *
       * @param    out
       *           The output destination.
       * @param    prevLink
       *           Link to the previous page. <code>null</code> for the
       *           first page of output.
       * @param    tocLink
       *           Link to the table of contents.
       * @param    nextLink
       *           Link to the next page. <code>null</code> for the
       *           last page of output.
       */
      private void writePageLinks(IndentingWriter out,
                                  String          prevLink,
                                  String          tocLink,
                                  String          nextLink)
      {
         String toc  = HtmlHelper.href(tocLink, "Table of Contents", false);
         
         writeCenterStart(out);
         writeTableOpen(out, 50, 0);
         writeRowStart(out, false, null);
         
         if (prevLink != null)
         {
            String prev = HtmlHelper.href(prevLink, "Previous Page", false);
            
            writeField(out,
                       prev, 
                       VALIGN_MIDDLE,
                       HALIGN_CENTER,
                       NORMAL,
                       true);
         }
         
         writeField(out,
                    toc, 
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    NORMAL,
                    true);
                    
         if (nextLink != null)
         {
            String next = HtmlHelper.href(nextLink, "Next Page", false);
            
            writeField(out,
                       next, 
                       VALIGN_MIDDLE,
                       HALIGN_CENTER,
                       NORMAL,
                       true);
         }
         
         writeRowEnd(out);
         writeTableClose(out);
         writeCenterEnd(out);         
      }
      
      /**
       * Writes the table header of a detailed report.
       *
       * @param    out
       *           The output destination.
       */
      private void writeDetailsTableHeader(IndentingWriter out)
      {
         writeTableOpen(out);
         
         writeRowStart(out, true, OLIVE);
         
         writeField(out,
                    "Filename", 
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "Line #",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "Column #",
                    VALIGN_MIDDLE, 
                    HALIGN_CENTER,
                    BOLD);
         writeField(out,
                    "Match Text",
                    VALIGN_MIDDLE,
                    HALIGN_CENTER,
                    BOLD);
                    
         writeRowEnd(out);
      }
   
      /**
       * Writes the table entry of a detailed report.
       *
       * @param    out
       *           The output destination.
       * @param    next
       *           The match for this entry.
       * @param    relPath
       *           The relative path to the "root" directory in which the
       *           source filenames exist.
       * @param    file
       *           The file in which this match exists.
       * @param    color
       *           <code>true</code> if this row should get a colored
       *           background.
       */
      private void writeDetailsTableEntry(IndentingWriter out,
                                          Match           next,
                                          String          relPath,
                                          String          file,
                                          boolean         color)
      {
         if (srcHelper == null)
         {
            throw new RuntimeException("Missing SourceFileHelper instance.");
         }
         
         String line = fmt.format(next.getLine());
         String col  = fmt.format(next.getColumn());
         
         // build the HREF to the cache file and line number
         String url = srcHelper.getURL(relPath, file, line);
         
         writeRowStart(out, color, null);
         
         writeField(out,
                    HtmlHelper.href(url, file, false), 
                    VALIGN_TOP,
                    HALIGN_LEFT,
                    NORMAL,
                    true);
         writeField(out,
                    line,
                    VALIGN_TOP,
                    HALIGN_RIGHT,
                    NORMAL);
         writeField(out,
                    col,
                    VALIGN_TOP,
                    HALIGN_RIGHT,
                    NORMAL);
         writeField(out,
                    next.getText(),
                    VALIGN_MIDDLE,
                    HALIGN_LEFT,
                    PREFORMAT);           
         
         writeRowEnd(out);
      }
      
      /**
       * Writes the condition matched for the given statistic.
       *
       * @param    out
       *           The output destination.
       * @param    stat
       *           The statistic to be reported upon.
       */
      private void writeCondition(IndentingWriter out, Statistic stat)
      {
         // generate a special footer describing the conditions that were
         // matched for this statistic
         Iterator      conds = stat.getConditionsIterator();
         StringBuilder sb    = new StringBuilder("Conditions matched: ");
         int           i     = 0;
         
         while (conds.hasNext())
         {
            String condition = HtmlHelper.filter((String) conds.next());
            
            // add our 'conjunction' if there is more than 1 condition
            if (i > 0)
            {
               sb.append(", ");
            }
            i++;
            
            sb.append("[ ").append(condition).append(" ]");
         }         
         
         write(out, sb.toString());
      }
      
      /**
       * Writes the body of a custom report based on the given input.
       *
       * @param    out
       *           The output destination.
       * @param    headers
       *           An ordered list of each column's label. This must be a
       *           list with elements of type <code>String</code>.  The
       *           number of elements will be the number of columns created
       *           for the table.
       * @param    rows
       *           An ordered list containing all row data.  This is a list
       *           which contains lists of each column's data for each row in
       *           the table.  This must be a list with elements that are
       *           lists with elements of type <code>String</code>.  The
       *           number of elements in each contained list must be the same
       *           as the number of elements of the <code>headers</code> list.
       */
      private void writeCustomReportDetails(IndentingWriter out,
                                            List            headers,
                                            List            rows)
      throws IllegalArgumentException
      {
         // only print a body if there are entries
         if (headers != null && !headers.isEmpty())
         {
            // track if this row's background needs to be colorized or not
            boolean color = false;
            
            // write table header
            writeTableOpen(out);         
            writeRowStart(out, true, OLIVE);
            
            Iterator hiter = headers.iterator();
            
            while (hiter.hasNext())
            {
               String label = (String) hiter.next();
               
               writeField(out,
                          label, 
                          VALIGN_MIDDLE,
                          HALIGN_CENTER,
                          BOLD);
            }
            
            writeRowEnd(out);
            
            // process all rows
            Iterator riter = rows.iterator();
            
            while (riter.hasNext())
            {
               List     next    = (List) riter.next();
               Iterator columns = next.iterator();
               boolean  first   = true;
               
               writeRowStart(out, color, null);
               
               // write each field/column
               while (columns.hasNext())
               {
                  String ftxt = (String) columns.next();
                  
                  writeField(out,
                             ftxt,
                             VALIGN_MIDDLE,
                             first ? HALIGN_LEFT : HALIGN_RIGHT,
                             NORMAL);
                  
                  if (first)
                     first = false;
               }
               
               writeRowEnd(out);
               
               // toggle this value for the next row
               color = !color;
            }
            
            writeTableClose(out);
         }
      }
      
      /**
       * Drives the writing of the header of a report, includes the standard
       * top of the document and then adds a single centered header element
       * to the body (with the title).
       *
       * @param    out
       *           The output destination.
       * @param    title
       *           The preformatted title text.
       */      
      private void writeHeader(IndentingWriter out, String title)
      {  
         // generate the top of the document
         writeSimpleHeader(out, title);
         
         writeHeading(out, 2, title);
      }
         
      /**
       * Drives the writing of a header level 2 sub-title in a single
       * centered header element.
       *
       * @param    out
       *           The output destination.
       * @param    level
       *           Heading level.
       * @param    title
       *           The preformatted title text.
       */      
      private void writeHeading(IndentingWriter out, int level, String title)
      {
         writeCenterStart(out);
         write(out, HtmlHelper.heading(level, title));
         writeCenterEnd(out);
      }
         
      /**
       * Start a center element.
       *
       * @param    out
       *           The output destination.
       */      
      private void writeCenterStart(IndentingWriter out)
      {
         write(out, HtmlHelper.startCenter());
         out.increment();
      }
         
      /**
       * Start a center element.
       *
       * @param    out
       *           The output destination.
       */      
      private void writeCenterEnd(IndentingWriter out)
      {
         out.decrement();
         write(out, HtmlHelper.endCenter());
      }
         
      /**
       * Write a table of contents with one entry per category. As a side
       * effect, the anchor name for each category will be calculated and
       * saved for subsequent emission.
       *
       * @param    out
       *           The output destination.
       */      
      private void writeTOC(IndentingWriter out)
      {
         if (categories != null)
         {
            Iterator cats = categories.keySet().iterator();
            
            // "header"
            writeSeparator(out);
            writeHeading(out, 3, "Table of Contents");
            write(out, HtmlHelper.startDiv());
            out.increment();
               
            // process all categories
            while (cats.hasNext())
            {
               String next = (String) cats.next();
               
               // create an anchor and a TOC entry that references it
               String anchor = writeTOCEntry(out, next);
               
               // save off the anchor name for this category (the value was
               // null before)
               categories.put(next, anchor);
            }
            
            // create an anchor and a TOC entry that references it
            allReportsAnchor = writeTOCEntry(out, ALL_RPTS);
            
            out.decrement();
            write(out, HtmlHelper.endDiv());
            writeSeparator(out);
         }
      }
      
      /**
       * Write a paragraph element and a following horizontal rule to the
       * given stream.
       *
       * @param    out
       *           The output destination.
       */
      private void writeSeparator(Writer out)
      {
         write(out, HtmlHelper.p());
         write(out, HtmlHelper.hr());
      }
      
      /**
       * Create an anchor and write an entry in the table of contents
       * which references that named anchor.
       *
       * @param    out
       *           The output destination.
       * @param    txt
       *           The description of the anchor location.
       *
       * @return   The anchor name calculated for the given entry.
       */      
      private String writeTOCEntry(IndentingWriter out, String txt)
      {  
         StringBuilder sb = new StringBuilder();
         String anchorTxt = generateSafeFilename("", txt, "");
         sb.append("#").append(anchorTxt);
         
         // create the link and insert it into the TOC
         write(out, HtmlHelper.href(sb.toString(), txt, false));
         write(out, HtmlHelper.br());
         
         return anchorTxt;
      }
      
      /**
       * Actual worker that writes the top of the header of a report, it does
       * not add any elements to the body although it sets the default
       * background color for the body.
       * <p>
       * This method implements HTML entity conversion to ensure that all
       * characters can be properly rendered.  The following subset are
       * converted:
       * <p>
       * &amp;
       * &quot;
       * &lt;
       * &gt;
       * &apos;
       *
       * @param    out
       *           The output destination.
       * @param    title
       *           The preformatted title text.
       */      
      private void writeSimpleHeader(IndentingWriter out, String title)
      {
         write(out, HtmlHelper.startHtml());
         out.increment();
         write(out, HtmlHelper.startHead());
         out.increment();
         write(out, HtmlHelper.getEncoding());
         write(out, HtmlHelper.title(title));
         out.decrement();
         write(out, HtmlHelper.endHead());
         write(out, HtmlHelper.startBody(WHITE));
         out.increment();
      }
         
      /**
       * Actual worker that writes the footer of a report which includes
       * some summary information.
       * <p>
       * This method implements HTML entity conversion to ensure that all
       * characters can be properly rendered.  The following subset are
       * converted:
       * <p>
       * &amp;
       * &quot;
       * &lt;
       * &gt;
       * &apos;
       *
       * @param    out
       *           The output destination.
       * @param    link
       *           <code>true</code> to enable generation of a source file
       *           index link.
       * @param    relPath
       *           The relative path to the "root" directory in which the
       *           source filenames exist. 
       */      
      private void writeFooter(IndentingWriter out,
                               boolean         link,
                               String          relPath)
      {
         if (link)
         {
            if (file_xref)
            {
               writeFileIndexLink(out, relPath);
            }
            else
            {
               write(out, describeAstList());         
            }
         }
         
         StringBuilder sb = new StringBuilder();
         
         // write timestamp
         sb.append("Generated at ")
           .append(sdf.format(new Date()))
           .append(".\n");
         
         // handle entity conversion using filter and add the text
         write(out, HtmlHelper.filter(sb.toString()));
         
         writeSimpleFooter(out);
      }
      
      /**
       * Write the common end of an HTML file.
       *
       * @param    out
       *           The output destination.
       */      
      private void writeSimpleFooter(IndentingWriter out)
      {
         // output the standard end of the file
         out.decrement();
         write(out, HtmlHelper.endBody());
         out.decrement();
         write(out, HtmlHelper.endHtml());
      }
         
      /**
       * Worker that writes a link to the source file index.
       *
       * @param    out
       *           The output destination.
       * @param    relPath
       *           The relative path to the "root" directory in which the
       *           output has been written. 
       */      
      private void writeFileIndexLink(IndentingWriter out, String relPath)
      {
         StringBuilder sb = new StringBuilder();
         sb.append(relPath).append(xref_name);
         
         // build link to the source file index
         String contents = HtmlHelper.href(sb.toString(),
                                           "Source File Index",
                                           false);
         
         write(out, "Please see the ");
         write(out, contents);
         write(out, " for the list of report inputs.  ");
      }
      
      /**
       * Create a string to describe the explicit list of ASTs or the AST
       * specifications used to create the input file list.
       *
       * @return   The list of filenames or file specifications. 
       */
      private String describeAstList() 
      {
         // document the file specifications for this report
         Iterator iter = resolver.getPatternEngine().astSpecs();
         
         StringBuilder sb = new StringBuilder("AST list generated using: ");
         int i = 0;
         
         while (iter.hasNext())
         {                     
            // add our 'conjunction' if there is more than 1 spec
            if (i > 0)
            {
               sb.append(", ");
            }
            i++;
            
            FileList fl = (FileList) iter.next();
            
            sb.append("< ");
            
            if (fl instanceof FileSpecList)
            {
               FileSpecList fsl      = (FileSpecList) fl;
               String       startDir = fsl.getStartDir().getAbsolutePath();
                              
               sb.append("&quot;").append(fsl.getFileSpec());
               sb.append("&quot; in ");
               sb.append(Configuration.normalizeFilename(startDir));
            }
            else
            {
               String[] list = fl.listFilenames();
               
               sb.append("List: ");
               
               for (int j = 0; j < list.length; j++)
               {
                  // add our 'conjunction' after the first name
                  if (j > 0)
                  {
                     sb.append(", ");
                  }
                  
                  sb.append(Configuration.normalizeFilename(list[j]));
               }
            }
            
            sb.append(" >");
         }         
         
         return sb.toString();
      }
      
      /**
       * Write the standard opening tags for a table.
       *
       * @param    out
       *           The output destination.
       */
      private void writeTableOpen(IndentingWriter out)
      {
         writeTableOpen(out, 100, 1);         
      }
   
      /**
       * Write the standard opening tags for a table.
       *
       * @param    out
       *           The output destination.
       * @param    width
       *           Percent of the page width the table should occupy.
       * @param    border
       *           The border thickness (0 for an invisible border).
       */
      private void writeTableOpen(IndentingWriter out,
                                  int             width,
                                  int             border)
      {         
         write(out, HtmlHelper.startTable(width, border));
         out.increment();
      }
   
      /**
       * Write the standard closing tags for a table.
       *
       * @param    out
       *           The output destination.
       */
      private void writeTableClose(IndentingWriter out)
      {
         out.decrement();         
         write(out, HtmlHelper.endTable());
         write(out, HtmlHelper.br());
      }
   
      /**
       * Write the opening row tag with optional background color support.
       * <p>
       * After opening the row, the indent level will be incremented.
       *
       * @param    out
       *           The output destination.
       * @param    bgcolor
       *           If <code>true</code>, the background of this row will not
       *           use the table or document background color.
       * @param    rgb
       *           A string starting with a '#' character and followed by 6
       *           hexidecimal digits specifying an RGB color.  If
       *           <code>null</code> this class' private constant GOLD will
       *           be used (which is RGB 255,204,102 or a light yellow).
       */
      private void writeRowStart(IndentingWriter out,
                                 boolean         bgcolor,
                                 String          rgb)
      {
         String contents = null;
         
         if (bgcolor)
         {
            contents = HtmlHelper.startRow(rgb == null ? GOLD : rgb);
         }
         else
         {
            contents = HtmlHelper.startRow();
         }
         
         write(out, contents);
         out.increment();
      }        
      
      /**
       * Write the closing row tag and decrement the indent level.
       *
       * @param    out
       *           The output destination.
       */
      private void writeRowEnd(IndentingWriter out)
      {
         out.decrement();
         write(out, HtmlHelper.endRow());
      }        
      
      /**
       * Write a single cell of text into the current row.  Alignment is
       * specified using one of the following:
       * <p>
       * <pre>
       * VALIGN_TOP
       * VALIGN_MIDDLE
       * VALIGN_BOTTOM
       * HALIGN_LEFT
       * HALIGN_CENTER
       * HALIGN_RIGHT
       * </pre>
       * <p>
       * This method forces HTML entity conversion to ensure that all
       * characters can be properly rendered.  The following subset are
       * converted at this time:
       * <p>
       * &amp;
       * &quot;
       * &lt;
       * &gt;
       * &apos;
       *
       * @param    out
       *           The output destination.
       * @param    text
       *           Contents of the cell.
       * @param    vAlign
       *           Vertical alignment of the cell.
       * @param    hAlign    
       *           Horizontal alignment of the cell.
       * @param    type
       *           Normal, Bold or Preformatted text if the type is set to
       *           <code>NORMAL, BOLD or PREFORMAT</code> respectively.
       */
      private void writeField(IndentingWriter out,
                              String          text, 
                              String          vAlign,
                              String          hAlign,
                              int             type)
      {
         writeField(out, text, vAlign, hAlign, type, false);
      }
      
      /**
       * Write a single cell of text into the current row.  Alignment is
       * specified using one of the following:
       * <p>
       * <pre>
       * VALIGN_TOP
       * VALIGN_MIDDLE
       * VALIGN_BOTTOM
       * HALIGN_LEFT
       * HALIGN_CENTER
       * HALIGN_RIGHT
       * </pre>
       * <p>
       * At the user's request, this method will implement HTML entity
       * conversion to ensure that all characters can be properly rendered.
       * The following subset are converted at this time:
       * <p>
       * &amp;
       * &quot;
       * &lt;
       * &gt;
       * &apos;
       *
       * @param    out
       *           The output destination.
       * @param    text
       *           Contents of the cell.
       * @param    vAlign
       *           Vertical alignment of the cell.
       * @param    hAlign    
       *           Horizontal alignment of the cell.
       * @param    type
       *           Normal, Bold or Preformatted text if the type is set to
       *           <code>NORMAL, BOLD or PREFORMAT</code> respectively.
       * @param    noFilter
       *           If <code>true</code> AND if the <code>type</code> is
       *           <code>NORMAL</code> or <code>PREFORMAT</code>, NO
       *           FILTERING of the text will be done.
       */
      private void writeField(IndentingWriter out,
                              String          text, 
                              String          vAlign,
                              String          hAlign,
                              int             type,
                              boolean         noFilter)
      {
         if (type == NORMAL)
         {
            // fast path
            write(out, HtmlHelper.field(hAlign, vAlign, text, !noFilter));
         }
         else
         {
            String start = HtmlHelper.startField(hAlign, vAlign);
            String end   = HtmlHelper.endField();
            
            if (type == BOLD)
            {
               write(out, start);
               out.increment();
               
               StringBuilder sb = new StringBuilder();
               sb.append(HtmlHelper.strong(text)).append("\n");
               
               write(out, sb.toString());
               
               out.decrement();
               write(out, end);
            }
            else if (type == PREFORMAT)
            {
               int level = out.getIndent();
               
               write(out, start);
               out.setIndent(0);
               
               text = noFilter ? text : HtmlHelper.filter(text);
               write(out, HtmlHelper.pre(text));
               
               out.setIndent(level);
               write(out, end);
            }
            else
            {
               String err = String.format("Unknown content type '%d'!", type);
               throw new RuntimeException(err);
            }
         }
      }
      
      /**
       * Handle output processing in a consistent manner by writing the
       * text to the given stream. Any failures will be transformed into
       * runtime exceptions.
       *
       * @param    out
       *           The output destination.
       * @param    text
       *           The content to write.
       *
       * @throws   RuntimeException
       *           On any failure.
       */
      private void write(Writer out, String text)
      {
         try
         {
            byte[] encoded = text.getBytes("UTF-8");
            String utf8 = new String(encoded, "UTF-8");
            out.write(utf8);
         }
         
         catch (IOException ioe)
         {
            throw new RuntimeException("I/O Failure on WRITE!", ioe);
         }
      }
      
      /**
       * Close the given writer and convert any failure into a runtime
       * exception.
       *
       * @param    out
       *           The output destination.
       *
       * @throws   RuntimeException
       *           On any failure.
       */
      private void close(Writer out)
      {
         try
         {
            out.close();
         }
         
         catch (IOException ioe)
         {
            throw new RuntimeException("I/O Failure on CLOSE!", ioe);
         }
      }
      
      /**
       * Creates a file-backed printable writer for storing report output.
       * As long as the filename is not a directory, it is deleted if it
       * exists and a new writer is created and returned. Note that the
       * user can specify an output directory.  If specified, this is
       * honored by prepending this (and the platform specific file
       * separator if necessary) to the name passed.
       * <p>
       * Any directories needed (that don't exist) will be created.
       *
       * @param    name
       *           Input file name.  May optionally include a relative path
       *           however it should not duplicate any path segments that
       *           would be created if <code>honor</code> is <code>true</code>
       *           since that is prepended independently.
       * @param    honor
       *           Determines if the output directory must be honored or
       *           not (<code>true</code> means prepend the output
       *           directory).
       *
       * @return   An open writer which to write the parser output.
       */
      private IndentingWriter createStream(String name, boolean honor)
      throws IllegalArgumentException
      {
         IndentingWriter out = null;
         StringBuilder   sb  = new StringBuilder();
         
         // prepend our output directory if we have one
         if (dir != null && honor)
         {            
            sb.append(dir);
         }
         
         // add our output filename
         sb.append(name);
         String fname = sb.toString();
         
         // create any directories needed
         Utils.createPathForTarget(fname);
         
         // build a file
         File file = new File(fname);
         
         // if this filename exists as a directory, we can't save results
         if (file.isDirectory())
         {   
            throw new IllegalArgumentException(name + " is a directory!");
         }
         
         // remove the target file if it exists
         if (file.exists())
         {
            file.delete();
         }
          
         try
         {
            FileOutputStream   fos = new FileOutputStream(file);
            OutputStreamWriter osw = new OutputStreamWriter(fos, "UTF-8");
            BufferedWriter     bw  = new BufferedWriter(osw);
            out = new IndentingWriter(bw);
         }
         catch (Exception exc)
         {
            throw new IllegalArgumentException("Cannot create stream for " +
                                               name + "(" + exc.toString() +
                                               ")!");
         }
         
         return out;             
      }
   }
   
   /**                                                                                  
    * Convert the given prefix into a value that refers to a directory.
    *
    * @param    prefix
    *           Directory prefix or <code>null</code>. If <code>null</code>
    *           then an empty string will be returned. Anything else will
    *           be normalized to ensure that it ends with a single file
    *           separator character and does NOT begin with a file separator
    *           character.
    *
    * @return   Safe prefix for use as a directory segment.
    */
   private String processDirPrefix(String prefix)
   {
      String pre;         
      
      if (prefix != null && prefix.length() > 0)
      {
         char sep = File.separatorChar;
         StringBuilder sb = new StringBuilder();
         
         // remove the separator from the front if needed
         sb.append(prefix.charAt(0) == sep ? prefix.substring(1)
                                           : prefix);
         
         // append the separator to the end if needed
         if (prefix.charAt(prefix.length() - 1) != File.separatorChar)
         {
            sb.append(sep);
         }
         
         pre = sb.toString();
      }
      else
      {
         // set to a safe value
         pre = "";
      }
      
      return pre;
   }
   
   /**
    * Calculate the number of directory levels in the given path and return
    * a string which represents the relative path from that point back to
    * the "root" of the given path.
    *
    * @param    path
    *           The path to analyze.
    *
    * @return   A string with a sequence of ".." segments, 1 for each level
    *           of the given <code>path</code>.
    */
   private String calcRelativePath(String path)
   {
      StringBuilder sb = new StringBuilder();
      
      int idx = path.indexOf(File.separatorChar);
      
      while (idx != -1)
      {
         sb.append("..").append(File.separatorChar);
         idx = path.indexOf(File.separatorChar, ++idx);
      }
      
      return sb.toString();
   }
   
   /**
    * Stores data associated with detailed report page processing.
    */
   private static class PageState
   {
      /** File name of the current file being processed. */
      private String file = null;
      
      /** List of matches for the given file. */
      private Iterator mset = null;
      
      /** File name of the first file processed on the last page. */
      private String start = null;
      
      /** File name of the last file processed on the last page. */
      private String end = null;
   }
   
   /**
    * Stores data associated with an entry in the list of reports (each
    * entry is the main report index or summary file).
    */
   private static class MasterReportEntry
   implements Serializable
   {
      /** Report description (same as the title of the report). */
      private String descr = null;
      
      /** Report URL (file name). */
      private String url = null;
      
      /** Total number of matches found in this report (across all files). */
      private int matches = 0;
      
      /** Total number of files found with at least 1 match. */
      private int files = 0;
      
      /** All categories with which this report is associated. */
      private Set categories = null;
   }
}

